# -*- coding: utf-8 -*-  pylint: disable=undefined-variable
'''
warfarin class
==============

This `warfarin` class implements a two compartment PK/PD model for warfarin.
'''

import functools
import itertools
from typing import Any, Dict, Iterable, List, Optional, Tuple

from reil.datatypes.feature import Feature, FeatureArray
from reil.healthcare.patient import Patient
from reil.healthcare.subjects.health_subject import HealthSubject
from reil.utils import reil_functions


class Warfarin(HealthSubject):
    '''
    A warfarin subject based on Hamberg's two compartment PK/PD model.
    '''

    def __init__(
            self,
            patient: Patient,
            INR_range: Tuple[float, float] = (0.0, 15.0),
            dose_range: Tuple[float, float] = (0.0, 15.0),
            interval_range: Tuple[int, int] = (1, 28),
            max_day: int = 90,
            **kwargs: Any):
        '''
        Arguments
        ---------
        patient:
            A patient object that generates new patients and models
            interaction between dose and INR.

        action_generator:
            An `ActionGenerator` object with 'dose' and
            'interval' components.

        max_day:
            Maximum duration of each trial.

        Raises
        ------
        ValueError
            action_generator should have a "dose" component.

        ValueError
            action_generator should have an "interval" component.
        '''

        super().__init__(
            patient=patient,
            measurement_name='INR',
            measurement_range=INR_range,
            dose_range=dose_range,
            interval_range=interval_range,
            max_day=max_day,
            **kwargs)

    def _generate_state_defs(self):
        super()._generate_state_defs()

        patient_basic: Tuple[Tuple[str, Dict[str, Any]], ...] = (
            ('age', {}), ('CYP2C9', {}),
            ('VKORC1', {}), ('sensitivity', {})
        )
        patient_extra: Tuple[Tuple[str, Dict[str, Any]], ...] = (
            ('weight', {}), ('height', {}),
            ('gender', {}), ('race', {}), ('tobaco', {}),
            ('amiodarone', {}), ('fluvastatin', {})
        )

        self.state.add_definition(
            'age', ('age', {}))

        self.state.add_definition(
            'patient_basic', *patient_basic)

        self.state.add_definition(
            'patient', *patient_basic, *patient_extra)

        self.state.add_definition(
            'patient_w_dosing',
            *patient_basic, *patient_extra,
            ('day', {}),
            ('dose_history', {'length': -1}),
            ('INR_history', {'length': -1}),
            ('interval_history', {'length': -1}))

        self.state.add_definition(
            'patient_for_baseline',
            *patient_basic, *patient_extra,
            ('day', {}),
            ('dose_history', {'length': 4}),
            ('INR_history', {'length': 4}),
            ('interval_history', {'length': 4}))

        for i in (1, 3, 5, 7, 9):
            self.state.add_definition(
                f'patient_w_dosing_{i:02}',
                *patient_basic,
                ('day', {}),
                ('dose_history', {'length': i}),
                ('INR_history', {'length': i}),
                ('interval_history', {'length': i}))

        self.state.add_definition(
            'patient_w_full_dosing',
            *patient_basic, *patient_extra,
            ('day', {}),
            ('daily_dose_history', {'length': -1}),
            ('daily_INR_history', {'length': -1}),
            ('interval_history', {'length': -1}))

        self.state.add_definition(
            'daily_INR',
            ('daily_INR_history', {'length': -1}))

        self.state.add_definition(
            'Measured_INR_2',
            ('INR_history', {'length': 2}),
            ('interval_history', {'length': 1}))

        self.state.add_definition(
            'INR_within_2',
            ('daily_INR_history', {'length': -1}))

    def _generate_reward_defs(self):
        super()._generate_reward_defs()

        reward_sq_dist = reil_functions.NormalizedSquareDistance(
            name='sq_dist', arguments=('daily_INR_history',),  # type: ignore
            length=-1, multiplier=-1.0, retrospective=True, interpolate=False,
            center=2.5, band_width=1.0, exclude_first=True)

        reward_sq_dist_interpolation = reil_functions.NormalizedSquareDistance(
            name='sq_dist_interpolation',
            arguments=('INR_history', 'interval_history'),  # type: ignore
            length=2, multiplier=-1.0, retrospective=True, interpolate=True,
            center=2.5, band_width=1.0, exclude_first=True)

        reward_PTTR = reil_functions.PercentInRange(
            name='PTTR', arguments=('daily_INR_history',),  # type: ignore
            length=-1, multiplier=-1.0, retrospective=True, interpolate=False,
            acceptable_range=(2, 3), exclude_first=True)

        reward_PTTR_interpolation = reil_functions.PercentInRange(
            name='PTTR',
            arguments=('INR_history', 'interval_history'),  # type: ignore
            length=2, multiplier=-1.0, retrospective=True, interpolate=True,
            acceptable_range=(2, 3), exclude_first=True)

        self.reward.add_definition(
            'sq_dist_exact', reward_sq_dist, 'INR_within_2')

        self.reward.add_definition(
            'sq_dist_interpolation', reward_sq_dist_interpolation,
            'Measured_INR_2')

        self.reward.add_definition(
            'PTTR_exact', reward_PTTR, 'INR_within_2')

        self.reward.add_definition(
            'PTTR_interpolation', reward_PTTR_interpolation, 'Measured_INR_2')

    def _generate_statistic_defs(self):
        super()._generate_statistic_defs()

        statistic_PTTR = reil_functions.PercentInRange(
            name='PTTR', arguments=('daily_INR_history',),  # type: ignore
            length=-1, multiplier=1.0, retrospective=True, interpolate=False,
            acceptable_range=(2, 3), exclude_first=True)

        self.statistic.add_definition(
            'PTTR_exact_basic', statistic_PTTR, 'daily_INR', 'patient_basic')

        self.statistic.add_definition(
            'PTTR_exact', statistic_PTTR, 'daily_INR', 'patient')

    def _generate_action_defs(self):
        super()._generate_action_defs()

        dose_gen = self.feature_gen_set['dose']
        interval_gen = self.feature_gen_set['interval']

        def _actions(
                dose_values: Iterable[float], interval_values: Iterable[int]):
            actions = itertools.product(
                (dose_gen(vi)
                 for vi in dose_values),
                (interval_gen(vi)
                 for vi in interval_values)
            )

            return tuple(FeatureArray(a) for a in actions)

        caps = tuple(i for i in (5.0, 10.0, 15.0)
                     if self._dose_range[0] <= i <= self._dose_range[1])
        dose = {cap: tuple(self.generate_dose_values(0.0, cap, 0.5))
                for cap in caps}

        int_fixed = {
            i: (i,) for i in (1, 2, 3, 7)
            if self._interval_range[0] <= i <= self._interval_range[1]}
        int_weekly = tuple(
            i for i in (7, 14, 21, 28)
            if self._interval_range[0] <= i <= self._interval_range[1])
        int_free = tuple(range(
            self._interval_range[0], self._interval_range[0] + 1))

        dose_int_fixed = {(d[0], i[0]): _actions(d[1], i[1])
                          for d, i in itertools.product(
                              dose.items(), int_fixed.items())
                          }

        dose_int_free = {k: _actions(v, int_free)
                         for k, v in dose.items()}

        dose_int_weekly = {k: _actions(v, int_weekly)
                           for k, v in dose.items()}

        max_cap = min(caps[-1], self._dose_range[1])

        def _237(f: FeatureArray, cap: float):
            day: int = f['day'].value  # type: ignore
            if day == 0:
                return dose_int_fixed[cap, 2]
            elif day == 2:
                return dose_int_fixed[max_cap, 3]
            elif day >= 5:
                return dose_int_fixed[max_cap, 7]
            else:
                raise ValueError(f'Wrong day: {day}.')

        for cap in caps:
            self.possible_actions.add_definition(
                f'daily_{int(cap):02}',
                lambda _: dose_int_fixed[cap, 1], 'day')

            self.possible_actions.add_definition(
                f'237_{int(cap):02}', functools.partial(_237, cap=cap), 'day')

            self.possible_actions.add_definition(
                f'free_{int(cap):02}', lambda _: dose_int_free[cap], 'day')

            self.possible_actions.add_definition(
                f'weekly_{int(cap):02}', lambda _: dose_int_weekly[cap], 'day')

    def _default_state_definition(
            self, _id: Optional[int] = None) -> FeatureArray:
        patient_features = self._patient.feature_set
        return FeatureArray([
            patient_features['age'],
            patient_features['CYP2C9'],
            patient_features['VKORC1']])

    def _sub_comp_age(self, _id: int, **kwargs: Any) -> Feature[float]:
        return super()._numerical_sub_comp('age')

    def _sub_comp_weight(self, _id: int, **kwargs: Any) -> Feature[float]:
        return self._numerical_sub_comp('weight')

    def _sub_comp_height(self, _id: int, **kwargs: Any) -> Feature[float]:
        return self._numerical_sub_comp('height')

    def _sub_comp_gender(self, _id: int, **kwargs: Any) -> Feature[str]:
        return self._categorical_sub_comp('gender')

    def _sub_comp_race(self, _id: int, **kwargs: Any) -> Feature[str]:
        return self._categorical_sub_comp('race')

    def _sub_comp_tobaco(self, _id: int, **kwargs: Any) -> Feature[str]:
        return self._categorical_sub_comp('tobaco')

    def _sub_comp_amiodarone(self, _id: int, **kwargs: Any) -> Feature[str]:
        return self._categorical_sub_comp('amiodarone')

    def _sub_comp_fluvastatin(self, _id: int, **kwargs: Any) -> Feature[str]:
        return self._categorical_sub_comp('fluvastatin')

    def _sub_comp_CYP2C9(self, _id: int, **kwargs: Any) -> Feature[str]:
        return self._categorical_sub_comp('CYP2C9')

    def _sub_comp_CYP2C9_masked(
            self, _id: int, days: int, **kwargs: Any) -> Feature[str]:
        return self._categorical_sub_comp('CYP2C9', self._day < days)

    def _sub_comp_VKORC1(self, _id: int, **kwargs: Any) -> Feature[str]:
        return self._categorical_sub_comp('VKORC1')

    def _sub_comp_VKORC1_masked(
            self, _id: int, days: int, **kwargs: Any) -> Feature[str]:
        return self._categorical_sub_comp('VKORC1', self._day < days)

    def _sub_comp_sensitivity(self, _id: int, **kwargs: Any) -> Feature[str]:
        return self._categorical_sub_comp('sensitivity')

    def _sub_comp_INR_history(
            self, _id: int, length: int = 1, **kwargs: Any
    ) -> Feature[List[float]]:
        return self._sub_comp_measurement_history(
            _id, length, **kwargs)

    def _sub_comp_daily_INR_history(
            self, _id: int, length: int = 1, **kwargs: Any
    ) -> Feature[List[float]]:
        return self._sub_comp_daily_measurement_history(
            _id, length, **kwargs)

    def _sub_comp_INR_within(
            self, _id: int, length: int = 1, **kwargs: Any
    ) -> Feature[List[float]]:
        intervals = self._get_history('interval_history', length).value
        return self._get_history('daily_INR', sum(intervals))  # type: ignore



# -*- coding: utf-8 -*-
'''
Intermountain class
===================

Intermountain warfarin dosing protocol based on `Anderson et al. (2007)
supplements Appendix B
<https://www.ahajournals.org/doi/10.1161/circulationaha.107.737312>`_
'''
# TODO: The current implementation assumes the same patient is being treated
# on future days. The correct implementation should be fully "functional", with
# no memory keeping! (no `additional_info`)
import functools
from typing import Any, Dict, List, Literal, Optional, Tuple

import reil.healthcare.dosing_protocols.dosing_protocol as dp


Zone = Literal['action point low',
               'red zone low',
               'yellow zone low',
               'green zone',
               'yellow zone high',
               'red zone high',
               'action point high']


class Intermountain(dp.DosingProtocol):
    '''
    Intermountain warfarin dosing protocol based on `Anderson et al. (2007)
    supplements Appendix B
    <https://www.ahajournals.org/doi/10.1161/circulationaha.107.737312>`_
    '''

    def __init__(self, enforce_day_ge_8: bool = True) -> None:
        super().__init__()
        self._enforce_day_ge_8 = enforce_day_ge_8

    def prescribe(self,
                  patient: Dict[str, Any],
                  additional_info: dp.AdditionalInfo
                  ) -> Tuple[dp.DosingDecision, dp.AdditionalInfo]:
        '''
        Prescribe a dose for the given `patient` and `additional_info`.

        Arguments
        ---------
        patient:
            A dictionary of patient characteristics including:
            - day
            - dose_history
            - interval_history (only for day 8)

        additional_info:
            A dictionary of information being communicated between protocols,
            including:
            - dosing_decisions_list
            - last_zone

        Returns
        -------
        :
            A `DosingDecision` along with updated `additional_info`.
        '''
        dosing_decisions_list: List[dp.DosingDecision] = additional_info.get(
            'dosing_decisions_list', [])
        last_zone: Zone = additional_info.get('last_zone', '')
        previous_INR = patient['INR_history'][-1]

        if not dosing_decisions_list:
            today = patient['day']

            if self._enforce_day_ge_8 and today < 8:
                raise ValueError('Intermountain is only valid for day>=8.')

            if self._enforce_day_ge_8 and today == 8:
                dose_history = patient['dose_history']
                interval_history = patient['interval_history']
                all_doses = functools.reduce(
                    lambda x, y: x+y,
                    ([dose_history[-i]]*interval_history[-i]
                     for i in range(len(interval_history), 0, -1)))

                if len(all_doses) < 3:
                    raise ValueError(
                        'Intermountain requires doses for days 5 to 7 '
                        'for dosing on day 8.')

                previous_dose = sum(all_doses[-3:])/3
            else:
                previous_dose = patient['dose_history'][-1]

            dosing_decisions_list, last_zone = self.intermountain_dosing_table(
                previous_INR, last_zone, previous_dose)

        else:
            if dosing_decisions_list[0].duration is None:
                dosing_decisions_list, last_zone = \
                    self.intermountain_dosing_table(
                        previous_INR, last_zone, dosing_decisions_list[0].dose)

        additional_info['last_zone'] = last_zone
        additional_info['dosing_decisions_list'] = dosing_decisions_list[1:]

        return dosing_decisions_list[0], additional_info

    @staticmethod
    def intermountain_dosing_table(  # noqa: C901
            INR: float,
            last_zone: Zone,
            daily_dose: float) -> Tuple[List[dp.DosingDecision], Zone]:
        '''
        Determine the dosing information, based on Intermountain dosing table.

        Arguments
        ---------
        current_INR:
            The latest value of INR.

        last_zone:
            The last zone the patient was in.
            * action point low
            * red zone low
            * yellow zone low
            * green zone
            * yellow zone high
            * red zone high
            * action point high

        daily_dose:
            The latest daily dose prescribed.

        Returns
        -------
        :
            * A list of `DoseInterval`s. It always includes the new daily dose
              and the new next test (in days). If an immediate dose is
              necessary, the first item will be the immediate dose and the next
              test day.
            * The new zone that patient's INR falls into.
        '''
        zone = Intermountain.zone(INR)

        weekly_dose = daily_dose * 7

        immediate_dose: float = -1.0
        immediate_duration: Optional[int] = None
        same_zone = zone == last_zone

        # -1s below are added to compensate for immediate_duration = 1
        next_duration: Optional[int] = {
            'action point low': (5-1, 14-1),
            'red zone low': (7-1, 14-1),
            'yellow zone low': (14, 14),
            'green zone': (14, 28),
            'yellow zone high': (14, 14),
            'red zone high': (7-1, 14-1),
            'action point high': (7, 14)
        }[zone][same_zone]

        if last_zone == 'action point high':
            if zone.startswith(('yellow', 'green')):
                weekly_dose *= 0.85
                next_duration = 7
            else:
                zone = 'action point high'
                immediate_dose = 0.0
                immediate_duration = 2
                next_duration = None

        elif zone == 'green zone':
            pass
        elif zone == 'action point low':
            # (immediate extra dose) average 5-7 for day 8
            immediate_dose = daily_dose * 2
            immediate_duration = 1
            weekly_dose *= 1.10
        elif zone == 'red zone low':
            # (extra half dose) average 5-7 for day 8
            immediate_dose = daily_dose * 1.5
            immediate_duration = 1
            weekly_dose *= 1.05
        elif zone == 'yellow zone low':
            if same_zone:
                weekly_dose *= 1.05
        elif zone == 'yellow zone high':
            if same_zone:
                weekly_dose *= 0.95
        elif zone == 'red zone high':
            immediate_dose = daily_dose * 0.5 if INR < 4.0 else 0.0
            immediate_duration = 1
            weekly_dose *= 0.90
        elif zone == 'action point high':
            immediate_dose = 0.0
            immediate_duration = 2
            next_duration = None

        dosing_decisions = list(
            (dp.DosingDecision(immediate_dose, immediate_duration),)
            if (immediate_dose >= 0.0 and immediate_duration is not None)
            else []
        ) + [dp.DosingDecision(weekly_dose / 7, next_duration)]

        return dosing_decisions, zone

    @staticmethod
    def zone(INR: float) -> Zone:
        '''
        Determine the zone based on patient's INR.

        Arguments
        ---------
        INR:
            the value of a patient's INR.

        Returns
        -------
        :
            Name of the dose, one of:
            * action point low
            * red zone low
            * yellow zone low
            * green zone
            * yellow zone high
            * red zone high
            * action point high
        '''
        if INR < 1.60:
            z = 'action point low'
        elif INR < 1.80:
            z = 'red zone low'
        elif INR < 2.00:
            z = 'yellow zone low'
        elif INR <= 3.00:
            z = 'green zone'
        elif INR < 3.40:
            z = 'yellow zone high'
        elif INR < 5.00:
            z = 'red zone high'
        else:
            z = 'action point high'

        return z



# -*- coding: utf-8 -*-
'''
IWPC class
==========

IWPC warfarin dosing protocols.

`clinical` and `pharmacogenetic` protocols based on `IWPC (2009)
<https://doi.org/10.1056/nejmoa0809329>`_

`modified` pharmacogenetic and `loading_dose` protocols based on
`Pirmohamed et al. (2013) <https://doi.org/10.1056/NEJMoa1311386>`_
'''

from math import exp
from typing import Any, Dict, Literal, Tuple

import reil.healthcare.dosing_protocols.dosing_protocol as dp
from reil.datatypes import feature


class IWPC(dp.DosingProtocol):
    '''
    IWPC warfarin dosing protocol, based on `IWPC (2009)
    <https://doi.org/10.1056/nejmoa0809329>`_
    '''

    def __init__(self,
                 method: Literal['pharmacogenetic',
                                 'clinical',
                                 'modified',
                                 'loading_dose'] = 'pharmacogenetic') -> None:
        '''
        Arguments
        ---------
        method:
            One of 'pharmacogenetic', 'clinical', 'modified', 'loading_dose'.
        '''
        if method.lower() == 'clinical':
            self._method = self.clinical
        elif method.lower() == 'modified':
            self._method = self.modified_pharmacogenetic
        elif method.lower() in ['pg', 'pharmacogenetic', 'default']:
            self._method = self.pharmacogenetic
        elif method.lower() in ['loading', 'loading_dose']:
            self._method = self.loading_dose
        else:
            raise ValueError(
                f'Unknown method {method}.\n'
                'Acceptable methods are pharmacogenetic, clinical, '
                'modified, and loading_dose.')

        self.reset()

    def prescribe(self,
                  patient: Dict[str, Any],
                  additional_info:  dp.AdditionalInfo
                  ) -> Tuple[dp.DosingDecision, dp.AdditionalInfo]:
        day = patient['day']

        if self._method is self.loading_dose:
            if day == 1:
                doses = self.loading_dose(patient)
                additional_info.update({'doses': doses})
            elif day <= 3:
                doses = additional_info['doses']
            else:
                raise ValueError(
                    'loading_dose method is only available for '
                    'days 1, 2, and 3.')

            dosing_decision = dp.DosingDecision(doses[day-1], 1)
        else:
            dosing_decision = dp.DosingDecision(
                self._method(patient))  # type: ignore
            additional_info.pop('doses', None)

        return dosing_decision, additional_info

    @staticmethod
    def clinical(patient: Dict[str, Any]) -> float:
        '''
        Determine warfarin dose using clinical IWPC formula.

        Arguments
        ---------
        patient:
            A dictionary of patient characteristics including:
            - age
            - height (in)
            - weight (lb)
            - race ('Asian', 'Black', 'Mixed', `MISSING`)
            - amiodarone ('Yes', 'No')
            - enzyme_inducer ('Yes' if patient takes any of
                              carbamazepine, phenytoin, rifampin, or
                              rifampicin, otherwise 'No')

        Returns
        -------
        :
            The maintenance dose
        '''
        age_in_decades = patient['age'] // 10
        height_in_cm = patient['height'] * 2.54
        weight_in_kg = patient['weight'] * 0.454

        race: Any = patient.get('race', feature.MISSING)
        african_american = race == 'Black'
        asian = race == 'Asian'
        missing_or_mixed_race = int(race in (feature.MISSING, 'Mixed'))
        amiodarone = patient.get('amiodarone') == 'Yes'
        enzyme_inducer = patient.get('enzyme_inducer') == 'Yes'

        weekly_dose = (4.0376
                       - 0.2546 * age_in_decades
                       + 0.0118 * height_in_cm
                       + 0.0134 * weight_in_kg
                       - 0.6752 * asian
                       + 0.4060 * african_american
                       + 0.0443 * missing_or_mixed_race
                       + 1.2799 * enzyme_inducer
                       - 0.5695 * amiodarone) ** 2

        return weekly_dose / 7.0

    @staticmethod
    def pharmacogenetic(patient: Dict[str, Any]) -> float:
        '''
        Determine warfarin dose using pharmacogenetic IWPC formula.

        Arguments
        ---------
        patient:
            A dictionary of patient characteristics including:
            - age
            - height (in)
            - weight (lb)
            - VKORC1 ('G/G', 'G/A', 'A/A', `MISSING`)
            - CYP2C9 ('*1/*1', '*1/*2', '*1/*3', '*2/*2', '*2/*3',
                      '*3/*3', `MISSING`)
            - race ('Asian', 'Black', 'Mixed', `MISSING`)
            - amiodarone ('Yes', 'No')
            - enzyme_inducer ('Yes' if patient takes any of
                              carbamazepine, phenytoin, rifampin, or
                              rifampicin, otherwise 'No')

        Returns
        -------
        :
            The maintenance dose
        '''
        age_in_decades = patient['age'] // 10
        height_in_cm = patient['height'] * 2.54
        weight_in_kg = patient['weight'] * 0.454

        vkorc1 = patient['VKORC1']
        vkorc1_ga = vkorc1 == 'G/A'
        vkorc1_aa = vkorc1 == 'A/A'
        vkorc1_unknown = vkorc1 == feature.MISSING

        cyp2c9 = patient['CYP2C9']
        cyp2c9_12 = cyp2c9 == '*1/*2'
        cyp2c9_13 = cyp2c9 == '*1/*3'
        cyp2c9_22 = cyp2c9 == '*2/*2'
        cyp2c9_23 = cyp2c9 == '*2/*3'
        cyp2c9_33 = cyp2c9 == '*3/*3'
        cyp2c9_unknown = cyp2c9 == feature.MISSING

        race: Any = patient.get('race', feature.MISSING)
        african_american = race == 'Black'
        asian = race == 'Asian'
        missing_or_mixed_race = int(race in (feature.MISSING, 'Mixed'))
        amiodarone = patient.get('amiodarone') == 'Yes'
        enzyme_inducer = patient.get('enzyme_inducer') == 'Yes'

        weekly_dose = (5.6044
                       - 0.2614 * age_in_decades
                       + 0.0087 * height_in_cm
                       + 0.0128 * weight_in_kg
                       - 0.8677 * vkorc1_ga
                       - 1.6974 * vkorc1_aa
                       - 0.4854 * vkorc1_unknown
                       - 0.5211 * cyp2c9_12
                       - 0.9357 * cyp2c9_13
                       - 1.0616 * cyp2c9_22
                       - 1.9206 * cyp2c9_23
                       - 2.3312 * cyp2c9_33
                       - 0.2188 * cyp2c9_unknown
                       - 0.1092 * asian
                       - 0.2760 * african_american
                       - 0.1032 * missing_or_mixed_race
                       + 1.1816 * enzyme_inducer
                       - 0.5503 * amiodarone) ** 2

        return weekly_dose / 7.0

    @staticmethod
    def modified_pharmacogenetic(patient: Dict[str, Any]) -> float:
        '''
        Determine warfarin dose using the modified pharmacogenetic IWPC
        formula.

        Arguments
        ---------
        patient:
            A dictionary of patient characteristics including:
            - age
            - height (in)
            - weight (lb)
            - VKORC1 ('G/G', 'G/A', 'A/A')
            - CYP2C9 ('*1/*1', '*1/*2', '*1/*3', '*2/*2', '*2/*3', '*3/*3')
            - amiodarone ('Yes', 'No')

        Returns
        -------
        :
            The maintenance dose
        '''
        age = patient['age']
        height_in_cm = patient['height'] * 2.54
        weight_in_kg = patient['weight'] * 0.454

        vkorc1 = patient['VKORC1']
        vkorc1_ga = vkorc1 == 'G/A'
        vkorc1_aa = vkorc1 == 'A/A'

        cyp2c9 = patient['CYP2C9']
        cyp2c9_12 = cyp2c9 == '*1/*2'
        cyp2c9_13 = cyp2c9 == '*1/*3'
        cyp2c9_22 = cyp2c9 == '*2/*2'
        cyp2c9_23 = cyp2c9 == '*2/*3'
        cyp2c9_33 = cyp2c9 == '*3/*3'

        amiodarone = patient.get('amiodarone') == 'Yes'

        weekly_dose = (5.6044
                       - 0.02614 * age
                       + 0.0087 * height_in_cm
                       + 0.0128 * weight_in_kg
                       - 0.8677 * vkorc1_ga
                       - 1.6974 * vkorc1_aa
                       - 0.5211 * cyp2c9_12
                       - 0.9357 * cyp2c9_13
                       - 1.0616 * cyp2c9_22
                       - 1.9206 * cyp2c9_23
                       - 2.3312 * cyp2c9_33
                       - 0.5503 * amiodarone) ** 2

        return weekly_dose / 7.0

    @staticmethod
    def loading_dose(patient: Dict[str, Any]) -> Tuple[float, float, float]:
        '''
        Determine warfarin dose using the modified pharmacogenetic IWPC
        formula.

        Arguments
        ---------
        patient:
            A dictionary of patient characteristics including:
            - age
            - height (in)
            - weight (lb)
            - VKORC1 ('G/G', 'G/A', 'A/A')
            - CYP2C9 ('*1/*1', '*1/*2', '*1/*3', '*2/*2', '*2/*3', '*3/*3')
            - amiodarone ('Yes', 'No')

        Returns
        -------
        :
            A list of warfarin doses for three days
        '''
        maintenance_dose = IWPC.modified_pharmacogenetic(patient)
        cyp2c9 = patient['CYP2C9']

        k24 = {'*1/*1': 0.0189,
               '*1/*2': 0.0158,
               '*1/*3': 0.0132,
               '*2/*2': 0.0130,
               '*2/*3': 0.0090,
               '*3/*3': 0.0075
               }[cyp2c9] * 24

        LD3 = maintenance_dose / ((1 - exp(-k24)) *
                                  (1 + exp(-k24) + exp(-2*k24)))
        # The following dose calculation is based on EU-PACT report page 19
        # Ravvaz uses the same formula, but uses weekly dose. However,
        # EU-PACT explicitly mentions "predicted daily dose (D)"
        LD3_WD = LD3 - maintenance_dose
        loading_doses = (
            maintenance_dose + LD3_WD * 1.5,
            maintenance_dose + LD3_WD,
            maintenance_dose + LD3_WD * 0.5)

        return loading_doses

    def reset(self) -> None:
        '''Reset the dosing protocol'''
        self._doses = []



import unittest

from reil.reilbase import ReilBase
from ruamel.yaml import YAML


class testReilBase(unittest.TestCase):
    def test_persistent(self):
        base_1 = ReilBase(name='test_persistent', version=1.0)
        base_1.save(filename='test_persistent')
        base_2 = ReilBase(name='base_2', version=0.6,
                          persistent_attributes=['name', 'version'])
        base_2.load(filename='test_persistent')

        self.assertEqual(base_2._name, 'base_2')
        self.assertEqual(base_2._version, 0.6)  # type: ignore

    def test_inheritance(self):
        class interited(ReilBase):
            def __init__(self, myarg: str = 'hello', **kwargs) -> None:
                self._myarg = myarg
                super().__init__(**kwargs)

        test = interited(another_arg=1, name='inherited')

        self.assertEqual(test._name, 'inherited')
        self.assertIn('_another_arg', test.__dict__)

    def test_yaml(self):
        config = '''
            base:
                reil.ReilBase:
                    name: test
                    path: .
        '''
        y = YAML().load(config)
        obj = ReilBase.parse_yaml(y['base'])
        self.assertIsInstance(obj, ReilBase)
        self.assertEqual(obj._name, 'test')  # type: ignore


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
'''
NoLearnAgent class
==================

The base class of all `agent` classes.
'''

import random
from typing import Any, Generator, Literal, Optional, Tuple, TypeVar, Union

from reil import stateful
from reil.datatypes.feature import FeatureArray

T = TypeVar('T')


class NoLearnAgent(stateful.Stateful):
    '''
    The base class of all `agent` classes. This class does not support any
    `learner`.
    '''

    def __init__(self,
                 default_actions: Tuple[FeatureArray, ...] = (),
                 tie_breaker: Literal['first', 'last', 'random'] = 'random',
                 **kwargs: Any):
        '''
        Arguments
        ---------
        default_actions:
            A list of default actions.

        tie_breaker:
            How to choose the `action` if more than one is candidate
            to be chosen.

        Raises
        ------
        ValueError:
            `tie_breaker` is not one of 'first', 'last', and 'random'.
        '''
        super().__init__(**kwargs)

        self._default_actions = default_actions
        self._training_trigger = 'none'

        if tie_breaker not in ['first', 'last', 'random']:
            raise ValueError(
                'Tie breaker should be one of first, last, or random options.')
        self._tie_breaker: Literal['first', 'last', 'random'] = tie_breaker

    def act(self,
            state: FeatureArray,
            subject_id: int,
            actions: Optional[Tuple[FeatureArray, ...]] = None,
            iteration: int = 0) -> FeatureArray:
        '''
        Return an action based on the given state.

        Arguments
        ---------
        state:
            The state for which the action should be returned.

        actions:
            The set of possible actions to choose from.

        iteration:
            The iteration in which the agent is acting.

        Returns
        -------
        :
            The action
        '''
        possible_actions = actions or self._default_actions

        result = self.best_actions(state, possible_actions)

        if len(result) > 1:
            action = self._break_tie(result, self._tie_breaker)
        else:
            action = result[0]

        return action

    def best_actions(self,
                     state: FeatureArray,
                     actions: Tuple[FeatureArray, ...],
                     ) -> Tuple[FeatureArray, ...]:
        '''
        Find the best `action`s for the given `state`.

        Arguments
        ---------
        state:
            The state for which the action should be returned.

        actions:
            The set of possible actions to choose from.

        Returns
        -------
        :
            A list of best actions.
        '''
        raise NotImplementedError

    def observe(self, subject_id: int, stat_name: str,
                ) -> Generator[Union[FeatureArray, None], Any, None]:
        '''
        Create a generator to interact with the subject (`subject_id`).

        This method creates a generator for `subject_id` that
        receives `state`, yields `action` and receives `reward`
        until it is closed. When `.close()` is called on the generator,
        `statistics` are calculated.

        Arguments
        ---------
        subject_id:
            the ID of the `subject` on which action happened.

        stat_name:
            The name of the `statistic` that should be computed at the end of
            each trajectory.

        Raises
        ------
        ValueError
            Subject with `subject_id` not found.
        '''
        if subject_id not in self._entity_list:
            raise ValueError(f'Subject with ID={subject_id} not found.')

        history: stateful.History = []
        new_observation = stateful.Observation()
        while True:
            try:
                new_observation = stateful.Observation()
                temp = yield
                new_observation.state = temp['state']
                actions: Tuple[FeatureArray, ...] = temp['actions']
                iteration: int = temp['iteration']

                if actions is not None:
                    new_observation.action = self.act(
                        state=new_observation.state,  # type: ignore
                        subject_id=subject_id,
                        actions=actions, iteration=iteration)

                    new_observation.reward = (yield new_observation.action)

                    history.append(new_observation)

            except GeneratorExit:
                if new_observation.reward is None:  # terminated early!
                    history.append(new_observation)

                # self._computed_stats.append(
                #     self.statistic(stat_name, subject_id))

                return

    @staticmethod
    def _break_tie(input_tuple: Tuple[T, ...],
                   method: Literal['first', 'last', 'random']) -> T:
        '''
        Choose one item from the supplied list of options, based on the method.

        Arguments
        ---------
        input_tuple:
            The set of options to choose from.

        method:
            Method of choosing an item from `input_tuple`.

        Returns
        -------
        :
            One of the items from the list


        :meta public:
        '''
        if method == 'first':
            action = input_tuple[0]
        elif method == 'last':
            action = input_tuple[-1]
        else:
            action = random.choice(input_tuple)

        return action



# -*- coding: utf-8 -*-
'''
subject class
=============

This `subject` class is the base class of all subject classes.
'''

import pathlib
from typing import Any, Optional, Tuple, Union

from reil import stateful
from reil.datatypes.components import SecondayComponent
from reil.datatypes.feature import Feature, FeatureArray


class Subject(stateful.Stateful):
    '''
    The base class of all subject classes.
    '''

    def __init__(self,
                 sequential_interaction: bool = True,
                 **kwargs: Any):
        '''
        Arguments
        ---------
        sequential_interaction:
            If `True`, `agents` can only act on the `subject` in the order they
            are added.

        Notes
        -----
        `sequential_interaction` is not enforced (implemented) yet!
        '''
        super().__init__(**kwargs)

        self._sequential_interaction = sequential_interaction
        self.reward = SecondayComponent(
            name='reward',
            primary_component=self.state,
            default_definition=self._default_reward_definition,
            enabled=False)

        self.possible_actions = SecondayComponent(
            name='action',
            primary_component=self.state,
            default_definition=self._default_action_definition,
            enabled=True)

    def _default_reward_definition(
            self, _id: Optional[int] = None) -> float:
        return 0.0

    def _default_action_definition(
            self, _id: Optional[int] = None) -> Tuple[FeatureArray, ...]:
        return (FeatureArray(Feature[Any](name='default_action')),)

    def is_terminated(self, _id: Optional[int] = None) -> bool:
        '''
        Determine if the `subject` is terminated for the given `agent` ID.

        Arguments
        ---------
        _id:
            ID of the agent that checks termination. In a multi-agent setting,
            e.g. an RTS game, one agent might die and another agent might still
            be alive.

        Returns
        -------
        :
            `False` as long as the subject can accept new actions from the
            `agent`. If `_id` is `None`, then returns `True` if no `agent`
            can act on the `subject`.
        '''
        raise NotImplementedError

    def take_effect(self, action: FeatureArray, _id: int = 0) -> None:
        '''
        Receive an `action` from `agent` with ID=`_id` and transition to
        the next state.

        Arguments
        ---------
        action:
            The action sent by the `agent` that will affect this `subject`.

        _id:
            ID of the `agent` that has sent the `action`.
        '''
        self.reward.enable()

    def reset(self) -> None:
        '''Reset the `subject`, so that it can resume accepting actions.'''
        super().reset()
        self.reward.disable()

    def load(self, filename: str,
             path: Optional[Union[str, pathlib.PurePath]]) -> None:

        super().load(filename, path=path)

        self.reward.set_primary_component(self.state)
        self.reward.set_default_definition(
            self._default_reward_definition)

    def save(self,
             filename: Optional[str] = None,
             path: Optional[Union[str, pathlib.PurePath]] = None,
             data_to_save: Optional[Tuple[str, ...]] = None
             ) -> Tuple[pathlib.PurePath, str]:

        prim_comp, self.reward._primary_component = (  # type: ignore
            self.reward._primary_component, None)
        reward_default, self.reward._default = (
            self.reward._default, None)
        try:
            f, p = super().save(filename, path=path, data_to_save=data_to_save)
        finally:
            self.reward._primary_component = prim_comp
            self.reward._default = reward_default

        return f, p



# -*- coding: utf-8 -*-
'''
SubjectDemon class
==================

`SubjectDemon` class changes the behavior of a given subject.
'''
from __future__ import annotations

import dataclasses
import pathlib
from typing import Any, Callable, Optional, Tuple, Union

from reil.datatypes.components import SecondayComponent, Statistic
from reil.datatypes.feature import FeatureArray
from reil.subjects.subject import Subject


@dataclasses.dataclass
class Modifier:
    name: str
    cond_state_def: Optional[str]
    condition_fn: Optional[Callable[[FeatureArray], bool]]
    modifier_fn: Callable[[FeatureArray], FeatureArray]

    def __post_init__(self):
        if self.condition_fn is not None:
            if self.cond_state_def is None:
                raise ValueError(
                    '`condition_fn` cannot be declared with '
                    '`cond_state_def=None`.')


class SubjectDemon(Subject):
    '''
    This class accepts a regular subject, and intervenes in its interaction
    with the agents. It can modify `state` representation or change
    the `possible_actions`.
    '''

    def __init__(
            self,
            subject: Optional[Subject] = None,
            action_modifier: Optional[Modifier] = None,
            state_modifier: Optional[Modifier] = None,
            **kwargs: Any):
        '''
        Arguments
        ---------
        subject:
            The `subject` that needs to be intervened with.

        action_modifier:
            A modifier instance for action.

        state_modifier:
            A modifier instance for state.

        '''
        super().__init__(**kwargs)

        self._subject: Subject
        self.reward: SecondayComponent
        self.statistic: Statistic

        if subject:
            self.__call__(subject)

        self._action_modifier = action_modifier
        self._state_modifier = state_modifier

    # @classmethod
    # def _empty_instance(cls):
    #     return cls(Subject())

    def __call__(self, subject: Subject) -> SubjectDemon:
        self._subject = subject
        self.reward = subject.reward
        self.statistic = subject.statistic
        self.is_terminated = subject.is_terminated
        self.take_effect = subject.take_effect
        self.reset = subject.reset

        return self

    def state(
            self, name: str,
            _id: Optional[int] = None) -> FeatureArray:
        '''
        Generate the component based on the specified `name` for the
        specified caller.

        Parameters
        ----------
        name:
            The name of the component definition.

        _id:
            ID of the caller.

        Returns
        -------
        :
            The component with the specified definition `name`.

        Raises
        ------
        ValueError
            Definition not found.
        '''
        original_state = self._subject.state(name, _id)
        modifier = self._state_modifier
        if (modifier is not None and
            (modifier.condition_fn is None
                or modifier.condition_fn(self._subject.state(
                    modifier.cond_state_def, _id)))):  # type: ignore
            return modifier.modifier_fn(original_state)

        return original_state

    def possible_actions(
            self, name: str,
            _id: Optional[int] = None) -> Any:
        '''
        Generate the component based on the specified `name` for the
        specified caller.

        Parameters
        ----------
        name:
            The name of the component definition.

        _id:
            ID of the caller.

        Returns
        -------
        :
            The component with the specified definition `name`.

        Raises
        ------
        ValueError
            Definition not found.
        '''
        original_set = self._subject.possible_actions(name, _id)
        modifier = self._action_modifier
        if (modifier is not None and
            (modifier.condition_fn is None
                or modifier.condition_fn(self._subject.state(
                    modifier.cond_state_def, _id)))):  # type: ignore
            return modifier.modifier_fn(original_set)

        return original_set

    def load(
            self, filename: str,
            path: Optional[Union[str, pathlib.PurePath]]) -> None:
        super().load(filename, path)

    def save(
            self,
            filename: Optional[str] = None,
            path: Optional[Union[str, pathlib.PurePath]] = None,
            data_to_save: Optional[Tuple[str, ...]] = None
    ) -> Tuple[pathlib.PurePath, str]:
        return super().save(filename, path, data_to_save)

    def register(self, entity_name: str, _id: Optional[int] = None) -> int:
        return self._subject.register(entity_name, _id)

    def deregister(self, entity_id: int) -> None:
        return self._subject.deregister(entity_id)



# -*- coding: utf-8 -*-
'''
agents module for reinforcement learning
========================================

This module provides different agents in reinforcement learning context.

Classes
-------
NoLearnAgent
    the base class of all agent classes

Agent
    the base class of all agent classes that learn from history

AgentDemon:
    A class that allows an `agent` to be manipulated.

QLearning
    the Q-learning agent that can accept any learner

DeepQLearning
    the agent with neural network as a learner (derived from
    `QLearning` class)

RandomAgent
    an agent that randomly chooses an action

UserAgent
    an agent that shows current state and asks for user's choice
    of action

Types
-----
TrainingData
    a type alias for training data, consisting of a tuple of
    `FeatureArray` for X matrix and a tuple of floats for Y vector.

'''
from .no_learn_agent import NoLearnAgent  # noqa: W601
from .agent import Agent, TrainingData  # noqa: W601
from .q_learning import QLearning  # noqa: W601
from .deep_q_learning import DeepQLearning  # noqa: W601
from .random_agent import RandomAgent  # noqa: W601
from .user_agent import UserAgent  # noqa: W601



# -*- coding: utf-8 -*-
'''
Utils module for reinforcement learning
=======================================

This module provides different utilities used in the `reil` package.

Submodules
----------
exploration_strategies:
    A module that provides different exploration strategies for `agents`.

functions:
    Contains different useful functions.

reil_functions:
    A set of special objects that are used to define and compute different
    statistics, rewards, etc.

Classes
-------
ActionGenerator:
    A class that accepts categorical and numerical components, and
    generates lists of actions as `FeatureArray` objects.

CategoricalComponent:
    A class used to define categorical parts of actions.

NumericalComponent:
    A class used to define numerical parts of actions.

InstanceGenerator:
    Accepts any object derived from `ReilBase`, and generates instances.

CommandlineArgument:
    A dataclass to define a commandline argument by specifying its name, type
    and default value.

CommandlineParser:
    A class to define and parse commandline arguments

ConfigParser:
    A class to read and parse `YAML` config files.

OutputWriter:
    A class to write statistics to a `csv` file.

MNKBoard:
    An m-by-n board in which k similar horizontal, vertical, or diagonal
    sequence is a win. Used in `subjects` such as `TicTacToe`.

WekaClusterer:
    A clustering class based on Weka's clustering capabilities (disabled)
'''
import reil.utils.exploration_strategies  # noqa: W0611
import reil.utils.functions  # noqa: W0611
import reil.utils.reil_functions  # noqa: W0611

from .action_generator import (ActionGenerator,  # noqa: W0611
                               CategoricalComponent, NumericalComponent)
from .instance_generator import InstanceGenerator  # noqa: W0611

from .argument_parser import (CommandlineArgument,   # noqa: W0611
                              CommandlineParser, ConfigParser)
from .output_writer import OutputWriter  # noqa: W0611

from .mnkboard import MNKBoard  # noqa: W0611



# -*- coding: utf-8 -*-
'''
VanillaExperienceReplay class
=============================

A `Buffer` with random pick that picks only if it is full.
'''

from typing import Dict, List, Optional, Tuple, Union

from reil.datatypes.buffers.circular_buffer import CircularBuffer
from reil.datatypes.buffers.buffer import T1, T2, PickModes


class VanillaExperienceReplay(CircularBuffer[T1, T2]):
    '''
    A `Buffer` with random pick that picks only if it is full.

    Extends `CircularBuffer` class.
    '''

    def __init__(self,
                 buffer_size: Optional[int] = None,
                 batch_size: Optional[int] = None,
                 buffer_names: Optional[List[str]] = None,
                 clear_buffer: bool = False) -> None:
        '''
        Initialize the buffer.

        Arguments
        -----------
        buffer_size:
            The size of the buffer.

        batch_size:
            The number of items to return at each `pick`.

        buffer_names:
            A list containing the names of buffer queues.

        clear_buffer:
            Whether to clear the buffer when `reset` is called.
        '''
        self._batch_size = None
        self._clear_buffer = False

        super().__init__(buffer_size=buffer_size,
                         buffer_names=buffer_names, pick_mode='random')

        self.setup(buffer_size=buffer_size,
                   batch_size=batch_size,
                   buffer_names=buffer_names,
                   clear_buffer=clear_buffer)

    def setup(
            self, buffer_size: Optional[int] = None,
            buffer_names: Optional[List[str]] = None,
            pick_mode: Optional[PickModes] = None,
            clear_buffer: Optional[bool] = None,
            batch_size: Optional[int] = None) -> None:
        '''
        Set up the buffer.

        Arguments
        ---------
        buffer_size:
            The size of the buffer.

        batch_size:
            The number of items to return at each `pick`.

        buffer_names:
            A list containing the names of buffer elements.

        clear_buffer:
            Whether to clear the buffer when `reset` is called.

        pick_mode:
            The default mode to pick items from the list. This argument is
            only available for signature consistency. Assigning it has
            no effect.

        Notes
        -----
        `setup` should be used only for attributes of the buffer that are
        not defined. Attempt to use `setup` to modify size, names or mode will
        result in an exception.
        '''
        super().setup(buffer_size=buffer_size,
                      buffer_names=buffer_names)

        if buffer_size is not None and buffer_size < 1:
            raise ValueError('buffer_size should be at least 1.')

        if self._buffer_size is not None:
            if batch_size is not None and self._buffer_size < batch_size:
                raise ValueError('buffer_size should be >= batch_size.')

        if batch_size is not None:
            if (self._buffer_size is not None and
                    self._buffer_size < batch_size):
                raise ValueError('buffer_size should be >= batch_size.')
            if self._batch_size not in (None, batch_size):
                raise ValueError(
                    'Cannot modify batch_size. The value is already set.')
            else:
                self._batch_size = batch_size

        self._clear_buffer = clear_buffer

    def pick(
        self,
        count: Optional[int] = None,
        mode: Optional[PickModes] = None
    ) -> Dict[str, Union[Tuple[T1, ...], Tuple[T2, ...]]]:
        '''
        Return items from the buffer.

        Arguments
        ---------
        count:
            This argument is only available for signature consistency.
            Assigning it has no effect.

        mode:
            This argument is only available for signature consistency.
            Assigning it has no effect.

        Returns
        -------
        :
            `batch_size` number of items from the buffer randomly.
            If the buffer is not full, return empty tuples.

        '''
        if self._buffer_full:
            return super().pick(self._batch_size, 'random')
        else:
            return {name: () for name in self._buffer}  # type: ignore

    def reset(self) -> None:
        '''
        Reset the buffer if `clear_buffer` is set to `True`.
        '''
        if self._clear_buffer:
            super().reset()
            self._buffer_full = False



# -*- coding: utf-8 -*-
'''
LookupTable and QLookupTable classes
====================================

`LookupTable` is a simple lookup table based on `dict` that checks for the
type of inputs to be `TableEntry`.

`QLookupTable` is a lookup table for `Q-learning`.
'''
import dataclasses
from typing import Any, Dict, Generic, Hashable, Tuple, TypeVar

from reil.learners.learner import Learner
from reil.datatypes.feature import FeatureArray
from reil.learners.learning_rate_schedulers import LearningRateScheduler

T = TypeVar('T')

# TODO: implement `load` and `save`


@dataclasses.dataclass
class TableEntry(Generic[T]):
    value: T
    N: int = 0


class LookupTable(Dict[Any, TableEntry[T]]):
    '''
    A simple lookup table based on `dict` that checks for the type of inputs
    to be `TableEntry`.
    '''
    def __setitem__(self, key: Hashable, item: Any):
        if isinstance(item, TableEntry):
            super().__setitem__(key, item)
        else:
            raise TypeError('item should be of type TableEntry.')


class QLookupTable(Learner[float]):
    '''
    A Q-learning lookup table class.

    This class stores input data and the corresponding output as a dictionay.
    '''
    def __init__(
            self,
            learning_rate: LearningRateScheduler,
            initial_estimate: float = 0.0,
            minimum_visits: int = 0) -> None:
        '''
        Arguments
        ---------
        learning_rate:
            A `LearningRateScheduler` object that determines the
        learning rate of this learner.

        initial_estimate:
            The value to be returned if not enough observations
        have been collected for a given x.

        minimum_visits:
            For a given input `x`, if it was learned for more than
        `minimum_visits`, the computed estimate is returned. For any less
        visited `x`, `initial_estimate` will be returned.
        '''
        self._learning_rate = learning_rate
        self._initial_estimate = initial_estimate
        self._minimum_visits = minimum_visits
        # defaultdict is not efficient.
        # It creates entries as soon as they are looked up.
        self._table: LookupTable[float] = LookupTable()

    def predict(self, X: Tuple[FeatureArray, ...]) -> Tuple[float, ...]:
        '''
        predict `y` for a given input list `X`.

        Arguments
        ---------
        X:
            A list of `FeatureArray` as inputs to the prediction model.

        Returns
        -------
        :
            The predicted `y`.
        '''
        dummy = TableEntry(self._initial_estimate)
        result = tuple(float(
            self._table.get(Xi, dummy).value
            if self._table.get(Xi, dummy).N >= self._minimum_visits
            else self._initial_estimate)
            for Xi in X)

        return result

    def learn(self, X: Tuple[FeatureArray, ...], Y: Tuple[float, ...]) -> None:
        '''
        Learn using the training set `X` and `Y`.

        Arguments
        ---------
        X:
            A list of `FeatureArray` as inputs to the learning model.

        Y:
            A list of float labels for the learning model.
        '''
        for i, Xi in enumerate(X):
            if Xi not in self._table:
                self._table[Xi] = TableEntry(self._initial_estimate)

            self._table[Xi].value += self._learning_rate.initial_lr * \
                (Y[i] - self._table[Xi].value)
            self._table[Xi].N += 1

    # def load(self, filename: str,
    #          path: Optional[Union[str, pathlib.PurePath]] = None) -> None:
    #     temp = defaultdict(
    #         lambda: {'value': self._initial_estimate,
    #                  'N': 0})
    #     _path = pathlib.Path(path if path is not None else '')
    #     with open(_path / f'{filename}.csv', 'r') as f:
    #         for k, v in csv.DictReader(f):
    #             temp[k] = v

    # def save(self,
    #          filename: str,
    #          path: Optional[Union[str, pathlib.PurePath]] = None
    #          ) -> Tuple[pathlib.PurePath, str]:
    #     _path = pathlib.Path(path if path is not None else '')
    #     with open(_path / f'{filename}.csv', 'w') as f:
    #         w = csv.writer(f)
    #         w.writerows(self._table.items())

    #     return _path, filename



# -*- coding: utf-8 -*-
'''
TicTacToe class
===============

The standard Tic-Tac-Toe game.
'''
import random
from typing import Any, Dict, Optional

from reil.subjects.mnkgame import MNKGame
from reil.datatypes.feature import FeatureArray, FeatureGenerator


class TicTacToe(MNKGame):
    '''
    Build a 3-by-3 board in which 2 players can play.
    Winner is the player who can put 3 pieces in one row, column, or diagonal.
    '''
    # _board is a row vector. (row, column) and index start from 0
    # _board_status: None: no winner yet,
    #                1..players: winner,
    #                0: stall,
    #               -1: illegal board

    def __init__(self, **kwargs: Any):
        super().__init__(m=3, n=3, k=3, players=2, **kwargs)
        self._state_gen = FeatureGenerator.numerical(
            name='state', lower=-1, upper=1)

    def default_state(self, _id: Optional[int] = None) -> FeatureArray:
        def modify(i: int, _id: Optional[int]) -> float:
            if i == _id:
                return 1
            if i == 0:
                return 0
            return -1

        return FeatureArray(self._state_gen(tuple(modify(i, _id)
                                                  for i in self._board)))

    def __repr__(self):
        return self.__class__.__qualname__


if __name__ == '__main__':
    board = TicTacToe()
    player: Dict[str, int] = {}
    p = 0
    player['P1'] = board.register('P1')
    player['P2'] = board.register('P2')
    while not board.is_terminated():
        board.state
        current_player = ['P1', 'P2'][p]
        print(p, current_player)
        actions = board.possible_actions(player[current_player])
        board.take_effect(random.choice(actions), player[current_player])
        print(f'{board}\n', board.reward(
            'default', player['P1']), board.reward('default', player['P2']))
        p = (p + 1) % 2



# -*- coding: utf-8 -*-
'''
EnvironmentStaticMap class
==========================

This class provides a learning environment for any reinforcement learning
`agent` on any `subject`. The interactions between `agents` and `subjects`
are determined by a fixed `interaction_sequence`.
'''
import pathlib
from typing import Any, Dict, List, NamedTuple, Optional, Tuple, Union

import pandas as pd
from reil.agents.agent_demon import AgentDemon
from reil.datatypes.interaction_protocol import InteractionProtocol
from reil.environments.environment import (EntityGenType, EntityType,
                                           Environment)
from reil.subjects.subject_demon import SubjectDemon


class StatInfo(NamedTuple):
    obj: str
    entity_name: str
    assigned_to: str
    a_s_name: Tuple[str, str]
    aggregators: Optional[Tuple[str, ...]]
    groupby: Optional[Tuple[str, ...]]


class EnvironmentStaticMap(Environment):
    '''
    Provide an interaction and learning environment for `agents` and
    `subjects`, based on a static interaction sequence.
    '''

    def __init__(
            self,
            entity_dict: Optional[Dict[str, Union[
                EntityType[Any], EntityGenType[Any], str]]] = None,
            demon_dict: Optional[Dict[str, Union[
                AgentDemon[Any], SubjectDemon, str]]] = None,
            interaction_sequence: Optional[
                Tuple[InteractionProtocol, ...]] = None,
            **kwargs: Any):
        '''
        Arguments
        ---------
        entity_dict:
            a dictionary that contains `agents`, `subjects`, and
            `generators`.

        interaction_sequence:
            a tuple of `InteractionProtocols` that specify
            how entities interact in the simulation.
        '''
        super().__init__(
            entity_dict=entity_dict, demon_dict=demon_dict, **kwargs)

        self._interaction_sequence: Tuple[InteractionProtocol, ...] = ()

        if interaction_sequence is not None:
            self.interaction_sequence = interaction_sequence

    def remove_entity(self, entity_names: Tuple[str, ...]) -> None:
        '''
        Extends `Environment.remove_entity`.

        Remove `agents`, `subjects`, or `instance_generators` from
        the environment.

        Arguments
        ---------
        entity_names:
            A list of `agent`/ `subject` names to be deleted.

        Raises
        ------
        RuntimeError
            The entity listed for deletion is used in the
            `interaction_sequence`.

        Notes
        -----
        This method removes the item from both `agents` and `subjects`
        lists. Hence, it is not recommended to use the same name for both
        an `agent` and a `subject`.
        '''
        names_in_use = [p.agent.name
                        for p in self._interaction_sequence] + \
                       [p.subject.name
                        for p in self._interaction_sequence]
        temp = set(entity_names).difference(names_in_use)
        if temp:
            raise RuntimeError(f'Some entities are in use: {temp}')

        super().remove_entity(entity_names)

    def remove_demon(self, demon_names: Tuple[str, ...]) -> None:
        '''
        Extends `Environment.remove_demon`.

        Remove `agent demons`, or `subject demons` from
        the environment.

        Arguments
        ---------
        demon_names:
            A list of `agent demon`/ `subject demon` names to be deleted.

        Raises
        ------
        RuntimeError
            The entity listed for deletion is used in the
            `interaction_sequence`.

        Notes
        -----
        This method removes the item from both `agent_demons` and
        `subject_demons` lists.
        Hence, it is not recommended to use the same name for both
        an `agent demon` and a `subject demon`.
        '''
        names_in_use = [p.agent.demon_name
                        for p in self._interaction_sequence] + \
                       [p.subject.demon_name
                        for p in self._interaction_sequence]
        temp = set(demon_names).difference(names_in_use)
        if temp:
            raise RuntimeError(f'Some demons are in use: {temp}')

        super().remove_demon(demon_names)

    @property
    def interaction_sequence(self) -> Tuple[InteractionProtocol, ...]:
        return self._interaction_sequence

    @interaction_sequence.setter
    def interaction_sequence(self,
                             seq: Tuple[InteractionProtocol, ...]) -> None:
        self._agent_observers = {}
        for protocol in seq:
            self.assert_protocol(protocol)
            self.register(protocol, get_agent_observer=True)

        self._interaction_sequence = seq

    def simulate_pass(self, n: int = 1) -> None:
        '''
        Go through the interaction sequence for a number of passes and
        simulate interactions accordingly.

        Arguments
        ---------
        n:
            The number of passes that simulation should go.
        '''
        for _ in range(n):
            for protocol in self._interaction_sequence:
                subject_name = protocol.subject.name

                if self._subjects[subject_name].is_terminated(None):
                    continue

                agent_name = protocol.agent.name
                a_s_name = (agent_name, subject_name)
                unit = protocol.unit
                state_name = protocol.state_name
                action_name = protocol.action_name
                reward_name = protocol.reward_name
                agent_id, _ = self._assignment_list[a_s_name]

                if protocol.subject.demon_name is None:
                    subject_instance = self._subjects[subject_name]
                else:
                    subject_instance = \
                        self._subject_demons[protocol.subject.demon_name](
                            self._subjects[subject_name])

                if unit == 'interaction':
                    self.interact(
                        agent_id=agent_id,  # type: ignore
                        agent_observer=self._agent_observers[a_s_name],
                        subject_instance=subject_instance,
                        state_name=state_name,
                        action_name=action_name,
                        reward_name=reward_name,
                        iteration=self._iterations[subject_name],
                        times=protocol.n)

                    if self._subjects[subject_name].is_terminated(None):
                        self.check_subject(subject_name)

                elif unit in ('instance', 'iteration'):
                    # For iteration, simulate the current instance, then in
                    # the next if statement, simulate the rest of the
                    # generated instances.
                    self.interact_while(
                        agent_id=agent_id,  # type: ignore
                        agent_observer=self._agent_observers[a_s_name],
                        subject_instance=subject_instance,
                        state_name=state_name,
                        action_name=action_name,
                        reward_name=reward_name,
                        iteration=self._iterations[subject_name])

                    if (unit == 'iteration'
                            and subject_name in self._instance_generators):
                        while self.check_subject(subject_name):
                            self.interact_while(
                                agent_id=agent_id,  # type: ignore
                                agent_observer=self._agent_observers[a_s_name],
                                subject_instance=subject_instance,
                                state_name=state_name,
                                action_name=action_name,
                                reward_name=reward_name,
                                iteration=self._iterations[subject_name])

                    else:
                        self.check_subject(subject_name)

                else:
                    raise ValueError(f'Unknown protocol unit: {unit}.')

    def simulate_to_termination(self) -> None:
        '''
        Go through the interaction sequence and simulate interactions
        accordingly, until all `subjects` are terminated.

        Notes
        -----
        To avoid possible infinite loops caused by normal `subjects`,
        this method is only available if all `subjects` are generated
        by `instance generators`.

        Raises
        ------
        TypeError:
            Attempt to call this method will normal subjects in the interaction
            sequence.
        '''
        subjects_in_use = set(s.subject.name
                              for s in self.interaction_sequence)
        no_generators = subjects_in_use.difference(self._instance_generators)
        if no_generators:
            raise TypeError(
                'Found subject(s) in the interaction_sequence that '
                f'are not instance generators: {no_generators}')

        infinites = [s
                     for s in subjects_in_use
                     if not self._instance_generators[s].is_finite]
        if infinites:
            raise TypeError('Found infinite instance generator(s) in the '
                            f'interaction_sequence: {infinites}')

        while not all(self._instance_generators[s].is_terminated()
                      for s in subjects_in_use):
            self.simulate_pass()

        self.report_statistics(True)

    def check_subject(self, subject_name: str) -> bool:
        '''
        Go over all `subjects`. If terminated, close related `agent_observers`,
        reset the `subject`, and create new `agent_observers`.
        '''
        # print(self._subjects[subject_name])
        affected_protocols = list(p for p in self._interaction_sequence
                                  if p.subject.name == subject_name)

        success: bool = True
        if affected_protocols:
            for p in affected_protocols:
                self.close_agent_observer(p)

            success = self.reset_subject(subject_name)

            for p in affected_protocols:
                self.register(p, get_agent_observer=True)

        return success

    def reset_subject(self, subject_name: str) -> bool:
        '''
        Extends `Environment.reset_subject()`.
        '''
        entities = set(
            (p.subject.statistic_name,
             self._assignment_list[(p.agent.name, p.subject.name)][1])
            for p in self.interaction_sequence
            if p.subject.name == subject_name and
            p.subject.statistic_name is not None)

        for e in entities:
            self._instance_generators.get(
                subject_name,
                self._subjects[subject_name]).statistic.append(*e)

        return super().reset_subject(subject_name)

    def report_statistics(self,
                          unstack: bool = True,
                          reset_history: bool = True
                          ) -> Dict[Tuple[str, str], pd.DataFrame]:
        '''Generate statistics for agents and subjects.

        Parameters
        ----------
        unstack:
            Whether to unstack the resulting pivottable or not.

        reset_history:
            Whether to clear up the history after computing stats.

        Returns
        -------
        :
            A dictionary with state-subject pairs as keys and dataframes as
            values.
        '''
        entities = set(
            StatInfo('_agents', p.agent.name, p.subject.name,
                     (p.agent.name, p.subject.name),
                     p.agent.aggregators, p.agent.groupby)
            for p in self.interaction_sequence
            if p.agent.statistic_name is not None)

        entities.update(set(
            StatInfo('_subjects', p.subject.name, p.agent.name,
                     (p.agent.name, p.subject.name),
                     p.subject.aggregators, p.subject.groupby)
            for p in self.interaction_sequence
            if p.subject.statistic_name is not None))

        def do_transform(x: pd.DataFrame) -> pd.DataFrame:
            return x.unstack().reset_index().rename(  # type: ignore
                columns={'level_0': 'aggregator', 0: 'value'})

        def no_transform(x: pd.DataFrame) -> pd.DataFrame:
            return x

        if unstack:
            transform = do_transform
        else:
            transform = no_transform

        result = {e.a_s_name:
                  transform(  # type: ignore
                      self._instance_generators.get(
                          e.entity_name,
                          self.__dict__[e.obj][e.entity_name]
                      ).statistic.aggregate(  # type: ignore
                          e.aggregators, e.groupby,
                          self._assignment_list[e.a_s_name][e.a_s_name.index(
                              e.entity_name)],
                          reset_history=reset_history)
                  ).assign(
                      entity=e.entity_name,
                      assigned_to=e.assigned_to,
                      iteration=self._iterations[e.a_s_name[1]])
                  for e in entities}

        return result

    def load(self,  # noqa: C901
             entity_name: Union[List[str], str] = 'all',
             filename: Optional[str] = None,
             path: Optional[Union[str, pathlib.PurePath]] = None) -> None:
        '''
        Load an entity or an `environment` from a file.

        Arguments
        ---------
        filename:
            The name of the file to be loaded.

        entity_name:
            If specified, that entity (`agent` or `subject`) is being
            loaded from file. 'all' loads an `environment`.

        Raises
        ------
        ValueError
            The filename is not specified.
        '''
        super().load(entity_name=entity_name, filename=filename, path=path)
        # To generate observers!
        self.interaction_sequence = self.interaction_sequence



# -*- coding: utf-8 -*-
'''
AgentDemon class
================

`AgentDemon` class changes the behavior of a given agent.
'''
from __future__ import annotations

import pathlib
from typing import Any, Callable, Optional, Tuple, Union

from reil import reilbase, stateful
from reil.agents.agent import Agent
from reil.agents.no_learn_agent import NoLearnAgent
from reil.datatypes.components import PrimaryComponent, Statistic
from reil.datatypes.feature import FeatureArray
from reil.learners.learner import LabelType


class AgentDemon(Agent[LabelType]):
    '''
    This class accepts a regular `agent`, and intervenes in its interaction
    with the subjects. A substitute `agent` acts whenever a condition is
    satisfied.
    '''

    def __init__(
            self,
            sub_agent: NoLearnAgent,
            condition_fn: Callable[[FeatureArray, int], bool],
            main_agent: Optional[Agent[LabelType]] = None,
            **kwargs: Any):
        '''
        Arguments
        ---------
        sub_agent:
            An `agent` that acts instead of the `main_agent`.

        condition_fn:
            A function that accepts the current state and ID of the subject
            and decides whether the `main_agent` should act or the `sub_agent`.

        main_agent:
            The `agent` that needs to be intervened with.
        '''
        reilbase.ReilBase.__init__(self, **kwargs)

        self.state: PrimaryComponent
        self.statistic: Statistic
        self._entity_list: stateful.EntityRegister
        self._training_trigger: str

        self._main_agent: Union[Agent[LabelType], None] = main_agent
        self._sub_agent: NoLearnAgent = sub_agent
        self._condition_fn = condition_fn

        if main_agent is not None:
            self.__call__(main_agent)

    @classmethod
    def _empty_instance(cls):
        return cls(NoLearnAgent(), lambda f, i: True, None)

    def __call__(self, main_agent: Agent[LabelType]) -> AgentDemon[LabelType]:
        self._main_agent = main_agent
        self.state = main_agent.state
        self.statistic = main_agent.statistic
        self._entity_list = main_agent._entity_list
        self._training_trigger = main_agent._training_trigger

        return self

    def load(self, filename: str,
             path: Optional[Union[str, pathlib.PurePath]]) -> None:
        _path = pathlib.Path(path or self._path)
        super().load(filename, _path)

        self._main_agent = (self._main_agent or Agent).from_pickle(
            filename, _path / 'main_agent')
        self._sub_agent = self._sub_agent.from_pickle(
            filename, _path / 'sub_agent')

        self.__call__(self._main_agent)

    def save(self,
             filename: Optional[str] = None,
             path: Optional[Union[str, pathlib.PurePath]] = None,
             data_to_save: Optional[Tuple[str, ...]] = None
             ) -> Tuple[pathlib.PurePath, str]:

        temp_main = self._main_agent
        temp_sub = self._sub_agent

        data = list(data_to_save or self.__dict__)
        save_main = '_main_agent' in data
        save_sub = '_sub_agent' in data
        if save_main:
            self._main_agent = type(self._main_agent)

        if save_sub:
            self._sub_agent = type(self._sub_agent)  # type: ignore

        if 'state' in data:
            data.remove('state')

        if 'statistic' in data:
            data.remove('statistic')

        _path = pathlib.Path(path or self._path)
        _filename = filename or self._name

        try:
            super().save(
                _filename, _path, data_to_save=tuple(data))
            if save_main:
                self._main_agent = temp_main
                if self._main_agent:
                    self._main_agent.save(_filename, _path / 'main_agent')
            if save_sub:
                self._sub_agent = temp_sub
                self._sub_agent.save(_filename, _path / 'sub_agent')
        finally:
            if save_main:
                self._main_agent = temp_main
                self._sub_agent = temp_sub

        return _path, _filename

    def register(self, entity_name: str, _id: Optional[int] = None) -> int:
        if self._main_agent is None:
            raise ValueError('main_agent is not set.')

        from_main = self._main_agent.register(entity_name, _id)
        from_sub = self._sub_agent.register(entity_name, _id)
        if from_main != from_sub:
            raise RuntimeError(f'ID from the main agent {from_main} does not '
                               f'match the ID from the sub agent {from_sub}.')

        return from_main

    def deregister(self, entity_id: int) -> None:
        self._sub_agent.deregister(entity_id)

        if self._main_agent is None:
            raise ValueError('main_agent is not set.')
        self._main_agent.deregister(entity_id)

    def reset(self):
        if self._main_agent:
            self._main_agent.reset()
        self._sub_agent.reset()

    def act(self,
            state: FeatureArray,
            subject_id: int,
            actions: Optional[Tuple[FeatureArray, ...]] = None,
            iteration: int = 0) -> FeatureArray:
        '''
        Return an action based on the given state.

        Arguments
        ---------
        state:
            the state for which the action should be returned.

        subject_id:
            the ID of the `subject` on which action should occur.

        actions:
            the set of possible actions to choose from.

        iteration:
            the iteration in which the agent is acting.

        Raises
        ------
        ValueError
            Subject with `subject_id` not found.

        Returns
        -------
        :
            the action
        '''
        if self._condition_fn(state, subject_id):
            return self._sub_agent.act(state, subject_id, actions, iteration)

        if self._main_agent is None:
            raise ValueError('main_agent is not set.')

        return self._main_agent.act(state, subject_id, actions, iteration)

    def learn(self, history: stateful.History) -> None:
        '''
        Learn using history.

        Arguments
        ---------
        subject_id:
            the ID of the `subject` whose history is being used for learning.

        next_state:
            The new `state` of the `subject` after taking `agent`'s action.
            Some methods
        '''
        if self._main_agent is None:
            raise ValueError('main_agent is not set.')

        self._main_agent.learn(history)



# -*- coding: utf-8 -*-
'''
Dense class
===========

The Dense learner.
'''
import pathlib
from typing import Any, List, Optional, Tuple, Union

import numpy as np
import tensorflow as tf
from reil.datatypes.feature import FeatureArray
from reil.learners.learner import Learner
from reil.learners.learning_rate_schedulers import (ConstantLearningRate,
                                                    LearningRateScheduler)
from tensorflow import keras


class Dense_tf_1(Learner[float]):
    '''
    The Dense learner.

    This class uses `tf.keras` to build a sequential dense network with one
    output.
    '''

    def __init__(
            self,
            learning_rate: LearningRateScheduler,
            validation_split: float = 0.3,
            hidden_layer_sizes: Tuple[int, ...] = (1,),
            input_length: Optional[int] = None,
            tensorboard_path: Optional[Union[str, pathlib.PurePath]] = None,
            **kwargs: Any) -> None:
        '''
        Arguments
        ---------
        learning_rate:
            A `LearningRateScheduler` object that determines the learning rate
            based on iteration. If any scheduler other than constant is
            provided, the model uses the `new_rate` method of the scheduler
            to determine the learning rate at each iteration.

        validation_split:
            How much of the training set should be used for validation?

        hidden_layer_sizes:
            A list of number of neurons for each layer.

        input_length:
            Size of the input data. If not supplied, the network will be
            generated based on the size of the first data point in `predict` or
            `learn` methods.

        tensorboard_path:
            A path to save tensorboard outputs. If not provided,
            tensorboard will be disabled.

        Raises
        ------
        ValueError
            Validation split not in the range of (0.0, 1.0).
        '''

        super().__init__(learning_rate=learning_rate, **kwargs)

        self._iteration: int = 0

        self._hidden_layer_sizes = hidden_layer_sizes
        self._input_length = input_length

        if not 0.0 < validation_split < 1.0:
            raise ValueError('validation split should be in (0.0, 1.0).')

        self._validation_split = validation_split

        self._callbacks: List[Any] = []
        self._tensorboard_path: Optional[pathlib.PurePath] = None
        self._model: keras.models.Sequential

        self._graph = tf.Graph()
        with self._graph.as_default():
            self._session = tf.Session()  # type: ignore

            self._model = keras.models.Sequential()

            if tensorboard_path is not None:
                self._tensorboard_path = pathlib.PurePath(
                    'logs', tensorboard_path)
                self._tensorboard = keras.callbacks.TensorBoard(
                    log_dir=self._tensorboard_path)
                # , histogram_freq=1)  #, write_images=True)
                self._callbacks.append(self._tensorboard)

            if not isinstance(self._learning_rate,
                              ConstantLearningRate):
                learning_rate_scheduler = \
                    keras.callbacks.LearningRateScheduler(
                        self._learning_rate.new_rate, verbose=0)
                self._callbacks.append(learning_rate_scheduler)

        self._ann_ready: bool = False
        if self._input_length is not None:
            self._generate_network()

    def _generate_network(self) -> None:
        '''
        Generate a multilayer neural net using `keras.Dense`.
        '''

        with self._graph.as_default():
            self._model = keras.models.Sequential()
            self._model.add(
                keras.layers.Dense(self._hidden_layer_sizes[0],
                                   activation='relu',
                                   name='layer_01',
                                   input_shape=(self._input_length,)))
            for i, v in enumerate(self._hidden_layer_sizes[1:]):
                self._model.add(keras.layers.Dense(
                    v, activation='relu', name=f'layer_{i+2:0>2}'))

            self._model.add(keras.layers.Dense(1, name='output'))

            self._model.compile(optimizer=keras.optimizers.Adam(
                learning_rate=self._learning_rate.initial_lr), loss='mae')

        self._ann_ready = True

    def predict(self, X: Tuple[FeatureArray, ...]) -> Tuple[float, ...]:
        '''
        predict `y` for a given input list `X`.

        Arguments
        ---------
        X:
            A list of `FeatureArray` as inputs to the prediction model.

        Returns
        -------
        :
            The predicted `y`.
        '''
        _X: List[List[Any]] = [x.normalized.flatten() for x in X]
        if not self._ann_ready:
            self._input_length = len(_X[0])
            self._generate_network()

        with self._session.as_default():
            with self._graph.as_default():
                result = self._model.predict(np.array(_X))  # type: ignore

        return result  # type: ignore

    def learn(self, X: Tuple[FeatureArray, ...], Y: Tuple[float, ...]) -> None:
        '''
        Learn using the training set `X` and `Y`.

        Arguments
        ---------
        X:
            A list of `FeatureArray` as inputs to the learning model.

        Y:
            A list of float labels for the learning model.
        '''
        _X: List[List[Any]] = [x.normalized.flatten() for x in X]
        if not self._ann_ready:
            self._input_length = len(_X[0])
            self._generate_network()

        with self._session.as_default():
            with self._graph.as_default():
                self._model.fit(
                    np.array(_X), np.array(Y),
                    initial_epoch=self._iteration, epochs=self._iteration+1,
                    callbacks=self._callbacks,
                    validation_split=self._validation_split,
                    verbose=0)

    def reset(self) -> None:
        '''
        reset the learner.
        '''
        self._iteration += 1

    def save(self,
             filename: Optional[str] = None,
             path: Optional[Union[str, pathlib.PurePath]] = None,
             data_to_save: Optional[Tuple[str, ...]] = None
             ) -> Tuple[pathlib.PurePath, str]:
        '''
        Extends `ReilBase.save` to handle `TF` objects.

        Arguments
        ---------
        filename:
            the name of the file to be saved.

        path:
            the path in which the file should be saved.

        data_to_save:
            This argument is only present for signature consistency. It has
            no effect on save.

        Returns
        -------
        :
            a `Path` object to the location of the saved file and its name
            as `str`
        '''
        p, f = super().save(
            filename, path,
            tuple(attr for attr in self.__dict__.keys()
                  if attr not in ('_session', '_graph', '_model')))

        with self._session.as_default():
            with self._graph.as_default():
                self._model.save(p / f'{f}.tf')

        return p, f

    def load(self,
             filename: str,
             path: Optional[Union[str, pathlib.PurePath]] = None) -> None:
        '''
        Extends `ReilBase.load` to handle `TF` objects.

        Arguments
        ---------
        filename:
            The name of the file to be loaded.

        path:
            Path of the location of the file.

        Raises
        ------
        ValueError:
            The filename is not specified.
        '''
        super().load(filename, path)

        _path = path or '.'
        self._graph = tf.Graph()  # type: ignore
        with self._graph.as_default():
            self._session = keras.backend.get_session()
            self._model = keras.models.load_model(
                pathlib.Path(_path, f'{filename}.tf'))

            self._tensorboard = keras.callbacks.TensorBoard(
                log_dir=self._tensorboard_path)

            if not isinstance(self._learning_rate,
                              ConstantLearningRate):
                learning_rate_scheduler = \
                    keras.callbacks.LearningRateScheduler(
                        self._learning_rate.new_rate, verbose=0)
                self._callbacks.append(learning_rate_scheduler)


class Dense_tf_2(Learner[float]):
    '''
    The Dense learner.

    This class uses `tf.keras` to build a sequential dense network with one
    output.
    '''

    def __init__(
            self,
            learning_rate: LearningRateScheduler,
            validation_split: float = 0.3,
            hidden_layer_sizes: Tuple[int, ...] = (1,),
            input_length: Optional[int] = None,
            tensorboard_path: Optional[Union[str, pathlib.PurePath]] = None,
            **kwargs: Any) -> None:
        '''
        Arguments
        ---------
        learning_rate:
            A `LearningRateScheduler` object that determines the learning rate
            based on iteration. If any scheduler other than constant is
            provided, the model uses the `new_rate` method of the scheduler
            to determine the learning rate at each iteration.

        validation_split:
            How much of the training set should be used for validation?

        hidden_layer_sizes:
            A list of number of neurons for each layer.

        input_length:
            Size of the input data. If not supplied, the network will be
            generated based on the size of the first data point in `predict` or
            `learn` methods.

        tensorboard_path:
            A path to save tensorboard outputs. If not provided,
            tensorboard will be disabled.

        Raises
        ------
        ValueError
            Validation split not in the range of (0.0, 1.0).
        '''

        super().__init__(learning_rate=learning_rate, **kwargs)

        self._iteration: int = 0

        self._hidden_layer_sizes = hidden_layer_sizes
        self._input_length = input_length

        if not 0.0 < validation_split < 1.0:
            raise ValueError('validation split should be in (0.0, 1.0).')

        self._validation_split = validation_split

        self._callbacks: List[Any] = []
        self._tensorboard_path: Optional[pathlib.PurePath] = None
        self._model = keras.models.Sequential()

        if tensorboard_path is not None:
            self._tensorboard_path = pathlib.PurePath(
                'logs', tensorboard_path)
            self._tensorboard = keras.callbacks.TensorBoard(
                log_dir=self._tensorboard_path)
            # , histogram_freq=1)  #, write_images=True)
            self._callbacks.append(self._tensorboard)

        if not isinstance(self._learning_rate,
                          ConstantLearningRate):
            learning_rate_scheduler = \
                keras.callbacks.LearningRateScheduler(
                    self._learning_rate.new_rate, verbose=0)
            self._callbacks.append(learning_rate_scheduler)

        self._ann_ready: bool = False
        if self._input_length is not None:
            self._generate_network()

    def _generate_network(self) -> None:
        '''
        Generate a multilayer neural net using `keras.Dense`.
        '''
        self._model = keras.models.Sequential()
        self._model.add(
            keras.layers.Dense(self._hidden_layer_sizes[0],
                               activation='relu',
                               name='layer_01',
                               input_shape=(self._input_length,)))
        for i, v in enumerate(self._hidden_layer_sizes[1:]):
            self._model.add(keras.layers.Dense(
                v, activation='relu', name=f'layer_{i+2:0>2}'))

        self._model.add(keras.layers.Dense(1, name='output'))

        self._model.compile(optimizer=keras.optimizers.Adam(
            learning_rate=self._learning_rate.initial_lr), loss='mae')

        self._ann_ready = True

    def predict(self, X: Tuple[FeatureArray, ...]) -> Tuple[float, ...]:
        '''
        predict `y` for a given input list `X`.

        Arguments
        ---------
        X:
            A list of `FeatureArray` as inputs to the prediction model.

        Returns
        -------
        :
            The predicted `y`.
        '''
        _X: List[List[Any]] = [x.normalized.flatten() for x in X]
        if not self._ann_ready:
            self._input_length = len(_X[0])
            self._generate_network()

        result = self._model.predict(np.array(_X))

        return result  # type: ignore

    def learn(self, X: Tuple[FeatureArray, ...], Y: Tuple[float, ...]) -> None:
        '''
        Learn using the training set `X` and `Y`.

        Arguments
        ---------
        X:
            A list of `FeatureArray` as inputs to the learning model.

        Y:
            A list of float labels for the learning model.
        '''
        _X: List[List[Any]] = [x.normalized.flatten() for x in X]
        if not self._ann_ready:
            self._input_length = len(_X[0])
            self._generate_network()

        self._model.fit(
            np.array(_X), np.array(Y),
            initial_epoch=self._iteration, epochs=self._iteration+1,
            callbacks=self._callbacks,
            validation_split=self._validation_split,
            verbose=0)

    def reset(self) -> None:
        '''
        reset the learner.
        '''
        self._iteration += 1

    def save(self,
             filename: Optional[str] = None,
             path: Optional[Union[str, pathlib.PurePath]] = None,
             data_to_save: Optional[Tuple[str, ...]] = None
             ) -> Tuple[pathlib.PurePath, str]:
        '''
        Extends `ReilBase.save` to handle `TF` objects.

        Arguments
        ---------
        filename:
            the name of the file to be saved.

        path:
            the path in which the file should be saved.

        data_to_save:
            This argument is only present for signature consistency. It has
            no effect on save.

        Returns
        -------
        :
            a `Path` object to the location of the saved file and its name
            as `str`
        '''
        p, f = super().save(
            filename, path,
            tuple(attr for attr in self.__dict__.keys()
                  if attr not in ('_session', '_graph', '_model')))

        try:
            self._model.save(pathlib.Path(p, f'{f}.tf').resolve())
        except ValueError:
            self._logger.warning(
                'Model is not compiled. Skipped saving the model.')

        return p, f

    def load(self,
             filename: str,
             path: Optional[Union[str, pathlib.PurePath]] = None) -> None:
        '''
        Extends `ReilBase.load` to handle `TF` objects.

        Arguments
        ---------
        filename:
            The name of the file to be loaded.

        path:
            Path of the location of the file.

        Raises
        ------
        ValueError:
            The filename is not specified.
        '''
        super().load(filename, path)

        _path = path or '.'
        if self._ann_ready:
            self._model = keras.models.load_model(
                pathlib.Path(_path, f'{filename}.tf'))
        else:
            self._model = keras.models.Sequential()

        self._tensorboard = keras.callbacks.TensorBoard(
            log_dir=self._tensorboard_path)

        if not isinstance(self._learning_rate,
                          ConstantLearningRate):
            learning_rate_scheduler = \
                keras.callbacks.LearningRateScheduler(
                    self._learning_rate.new_rate, verbose=0)
            self._callbacks.append(learning_rate_scheduler)


if tf.__version__[0] == '1':  # type: ignore
    Dense = Dense_tf_1
else:
    Dense = Dense_tf_2



# -*- coding: utf-8 -*-
'''
MNKBoard class
==============

This class creates a board for players to play the mnk game.
'''

import math
from typing import Any, Iterator, List, Optional, Tuple, Union, overload

from reil.datatypes.feature import FeatureArray, FeatureGenerator


class MNKBoard:
    '''
    Provide an m-by-n board to play.
    '''

    def __init__(self, m: int = 3, n: int = 3, k: int = 3, players: int = 2,
                 can_recapture: bool = True, **kwargs: Any):
        '''
        Arguments
        ---------
        m:
            The number of rows.

        n:
            The number of columns.

        k:
            The winning criterion, i.e. the number of cells in one row, column,
            or diagonal that a player needs to capture to win the game.

        players:
            The number of players.

        can_recapture:
            Whether a piece can be put on an occupied square
        '''
        self._m = m
        self._n = n
        self._k = k
        self._players = players
        self._can_recapture = can_recapture
        self._board_state_gen = FeatureGenerator.numerical(
            name='state', lower=0, upper=self._players)
        self.reset()

    @overload
    def set_piece(self, player: int, index: int) -> None:
        '''
        Set a piece for a player.

        Arguments
        ---------
        player:
            ID of the player whose piece will be set on the board.

        index:
            Where to put the piece. Index starts from 0 and assumes the board
            to be a list.

        Raises
        ------
        ValueError:
            Wrong player ID.

        ValueError:
            index is out of range.
        '''
        ...

    @overload
    def set_piece(self, player: int, row: int, column: int) -> None:
        '''
        Set a piece for a player.

        Arguments
        ---------
        player:
            ID of the player whose piece will be set on the board.

        row:
            The row on which the piece should be placed.

        column:
            The column on which the piece should be placed.

        Raises
        ------
        ValueError:
            Wrong player ID.

        ValueError:
            index is out of range.
        '''
        ...

    def set_piece(  # type: ignore
            self, player: int, index: Optional[int] = None,
            row: Optional[int] = None, column: Optional[int] = None
            ) -> None:
        '''
        Set a piece for a player.

        Arguments
        ---------
        player:
            ID of the player whose piece will be set on the board.

        row:
            The row on which the piece should be placed.

        column:
            The column on which the piece should be placed.

        index:
            Where to put the piece. Index starts from 0 and assumes the board
            to be a list.

        Raises
        ------
        ValueError:
            Wrong player ID.

        ValueError:
            index is out of range.

        Notes
        -----
        Either `index` or `row` and `column` should be used. If both are used,
        `row` and `column` is used.
        '''
        if player <= 0 or player > self._players:
            raise ValueError('player not found.')

        if row is None or column is None:
            if index is None:
                raise ValueError('No row-column pair or index found.')
            else:
                _index = index
        else:
            _index = row * self._n + column

        if self._board[_index] and not self._can_recapture:
            raise ValueError('The square is already occupied.')
        self._board[_index] = player

    @overload
    def clear_square(self, index: int) -> None:
        '''
        Set a piece for a player.

        Arguments
        ---------
        index:
            index to be cleared. Index starts from 0 and assumes the board
            to be a list.

        Raises
        ------
        ValueError:
            index is out of range.
        '''
        ...

    @overload
    def clear_square(self, row: int, column: int) -> None:
        '''
        Set a piece for a player.

        Arguments
        ---------
        row:
            The row on which a cell should be cleared.

        column:
            The column on which a cell should be cleared.

        Raises
        ------
        ValueError:
            index is out of range.
        '''
        ...

    def clear_square(  # type: ignore
            self, index: Optional[int] = None,
            row: Optional[int] = None, column: Optional[int] = None
            ) -> None:
        '''
        Set a piece for a player.

        Arguments
        ---------
        row:
            The row on which a cell should be cleared.

        column:
            The column on which a cell should be cleared.

        index:
            index to be cleared. Index starts from 0 and assumes the board
            to be a list.

        Raises
        ------
        ValueError:
            index is out of range.

        Notes
        -----
        Either `index` or `row` and `column` should be used. If both are used,
        `row` and `column` is used.
        '''

        if row is None or column is None:
            if index is None:
                raise ValueError('No row-column pair or index found.')
            else:
                _index = index
        else:
            _index = row * self._n + column

        self._board[_index] = 0

    @property
    def board_state(self):
        ''' Return the state of the board as a FeatureArray.'''
        return FeatureArray(self._board_state_gen(tuple(self._board)))

    def get_board(self,
                  as_list: bool = True) -> Union[List[int], List[List[int]]]:
        '''
        Return the board.

        Arguments
        ---------
        as_list:
            Whether to return the board as a list or a matrix.

        Returns
        -------
        :
            The board as a list or a 2D matrix.
        '''
        if as_list:
            return self._board
        else:
            return self.list_to_matrix(self._board, self._m, self._n)

    def get_action_set(self, as_list: bool = True
                       ) -> Union[Iterator[int], Iterator[Tuple[int, int]]]:
        '''
        Return a list of indexes of empty squares.

        Arguments
        ---------
        as_list:
            Whether to return the board as a list or a matrix.
        '''
        index = (i for i in range(self._m*self._n) if self._board[i] == 0)
        for action in index:
            if as_list:
                yield action
            else:
                yield (action // self._n, action % self._n)

    def reset(self) -> None:
        '''Clear the board.'''
        self._board = [0]*(self._m*self._n)

    def __str__(self):
        '''Return a printable format string of the board.'''
        return ('\n'.join(
            [''.join([f'{item:4}' for item in row])
             for row in self.list_to_matrix(self._board, self._m, self._n)]))

    @staticmethod
    def list_to_matrix(board: List[Any],
                       m: int, n: Optional[int] = None) -> List[List[Any]]:
        '''
        Covert a list to a 2D matrix.

        Arguments
        ---------
        board:
            The board to be converted.

        m:
            The number of rows.

        n:
            The number of columns. If omitted, it will be infered from the
            `board` size and `m`.
        '''
        _n = n or math.ceil(len(board) / m)

        return [board[row*_n:(row+1)*_n] for row in range(m)]

    def __repr__(self):
        return (self.__class__.__qualname__ + f', {self._m} x {self._n} board,'
                f' {self._k} on a line wins, {self._players} players')


if __name__ == '__main__':
    # create a board and set piece for each player and print the board
    board = MNKBoard(m=3, n=3, k=3, players=3)
    board.set_piece(1, row=0, column=0)
    board.set_piece(2, index=4)
    board.set_piece(3, index=8)
    print(f'{board}')



import setuptools
import versioneer

with open('README.rst') as f:
    long_description = f.read()

with open("requirements.txt", "r") as fh:
    requirements = [line.strip() for line in fh]

setuptools.setup(name='ReiL',
                 version=versioneer.get_version(),
                 cmdclass=versioneer.get_cmdclass(),
                 description='A Reinforcement Learning Module for Python',
                 long_description=long_description,
                 long_description_content_type="text/x-rst",
                 url='https://research-git.uiowa.edu/sanzabizadeh/Reinforcement-Learning',
                 author='Sadjad Anzabi Zadeh',
                 author_email='sadjad-anzabizadeh@uiowa.edu',
                 license='MIT License',
                 packages=setuptools.find_packages(exclude=('tests', 'docs')),
                 classifiers=[
                     "Programming Language :: Python :: 3",
                     "License :: OSI Approved :: MIT License",
                     "Operating System :: OS Independent",
                 ],
                 python_requires='>=3.8',
                 install_requires=requirements,
                 )



# -*- coding: utf-8 -*-
'''
Stateful class
==============

The base class of all stateful classes in `reil` package.

Methods
-------
state:
    the state of the entity (`agent` or `subject`) as a FeatureArray. Different
    state definitions can be introduced using `state.add_definition` method.
    _id is available, in case in the implementation, state is caller-dependent.
    (For example in games with partial map visibility).
    For subjects that are turn-based, it is a good practice to check
    that an agent is retrieving the state only when it is the agent's
    turn.

statistic:
    compute the value of the given statistic for the entity `_id`
    based on the statistic definition `name`. It should normally be called
    after each sampled path (trajectory). Different statistic definitions
    can be introduced using `statistic.add_definition` method.

_extract_sub_components:
    Extract methods that begin with `_sub_comp_`.

register:
    Register an external `entity` (`agents` for `subjects` and vice versa.)

deregister:
    Deregister an external `entity` (`agents` for `subjects` and vice versa.)
'''

from __future__ import annotations

import dataclasses
import pathlib
from typing import Any, Dict, List, Optional, Tuple, Union

from reil import reilbase
from reil.datatypes.components import (PrimaryComponent, Statistic,
                                       SubComponentInfo)
from reil.datatypes.feature import Feature, FeatureArray


@dataclasses.dataclass
class Observation:
    state: Optional[FeatureArray] = None
    action: Optional[FeatureArray] = None
    reward: Optional[float] = None


History = List[Observation]


class EntityRegister:
    '''
    Create and maintain a list of registered `entities`.


    :meta private:
    '''

    def __init__(self, min_entity_count: int, max_entity_count: int,
                 unique_entities: bool = True):
        '''
        Arguments
        ---------
        min_entity_count:
            The minimum number of `entities` needed to be registered so that
            the current `instance` is ready for interaction.

        max_entity_count:
            The maximum number of `entities` that can interact with this
            instance.

        unique_entities:
            If `True`, each `entity` can be registered only once.
        '''
        self._id_list: List[int] = []
        self._entity_list: List[str] = []
        self._min_entity_count = min_entity_count
        self._max_entity_count = max_entity_count
        self._unique_entities = unique_entities

    @property
    def ready(self) -> bool:
        '''
        Determine if enough `entities` are registered.

        Returns
        -------
        :
            `True` if enough `entities` are registered, else `False`.
        '''
        return len(self._id_list) >= self._min_entity_count

    def append(self, entity_name: str, _id: Optional[int] = None) -> int:
        '''
        Add a new `entity` to the end of the list.

        Parameters
        ----------
        entity_name:
            The name of the `entity` to add.

        _id:
            If provided, method tries to register the `entity` with the given
            ID.

        Returns
        -------
        :
            The ID assigned to the `entity`.

        Raises
        ------
        ValueError:
            Capacity is reached. No new `entities` can be registered.

        ValueError:
            ID is already taken.

        ValueError:
            `entity_name` is already registered with a different ID.
        '''
        if (0 < self._max_entity_count < len(self._id_list)):
            raise ValueError('Capacity is reached. No new entities can be'
                             ' registered.')

        # new_id = int(_id)  # type: ignore
        if self._unique_entities:
            if entity_name in self._entity_list:
                current_id = self._id_list[
                    self._entity_list.index(entity_name)]
                if _id is None or _id == current_id:
                    return current_id
                else:
                    raise ValueError(
                        f'{entity_name} is already registered with '
                        f'ID: {current_id}.')
            # elif _id is None:
            #     new_id = max(self._id_list, default=0) + 1
            elif _id in self._id_list:
                raise ValueError(f'{_id} is already taken.')
            # else:
            #     new_id = _id
        # elif _id is None:
        #     new_id = max(self._id_list, default=0) + 1
        elif _id in self._id_list:
            current_entity = self._entity_list[
                self._id_list.index(_id)]
            if entity_name == current_entity:
                return _id
            else:
                raise ValueError(f'{_id} is already taken.')
        # else:
        #     new_id = _id

        new_id = _id or max(self._id_list, default=0) + 1

        self._entity_list.append(entity_name)
        self._id_list.append(new_id)

        return new_id

    def remove(self, _id: int):
        '''
        Remove the `entity` registered by ID=`_id`.

        Arguments
        ---------
        _id:
            ID of the `entity` to remove.
        '''
        entity_name = self._entity_list[self._id_list.index(_id)]
        self._entity_list.remove(entity_name)
        self._id_list.remove(_id)

    def clear(self):
        '''
        Clear the list.
        '''
        self._id_list: List[int] = []
        self._entity_list: List[str] = []

    def __contains__(self, _id: int) -> bool:
        return _id in self._id_list


class Stateful(reilbase.ReilBase):
    '''
    The base class of all stateful classes in the `ReiL` package.
    '''

    def __init__(self,
                 min_entity_count: int = 1,
                 max_entity_count: int = -1,
                 unique_entities: bool = True,
                 **kwargs: Any):

        super().__init__(**kwargs)

        self.sub_comp_list = self._extract_sub_components()
        self.state = PrimaryComponent(self,
                                      self.sub_comp_list,
                                      self._default_state_definition)
        self.statistic = Statistic(
            name='statistic', primary_component=self.state,
            default_definition=self._default_statistic_definition)

        self._entity_list = EntityRegister(min_entity_count=min_entity_count,
                                           max_entity_count=max_entity_count,
                                           unique_entities=unique_entities)

    def _default_state_definition(
            self, _id: Optional[int] = None) -> FeatureArray:
        return FeatureArray(Feature[Any](name='default_state'))

    def _default_statistic_definition(
            self, _id: Optional[int] = None) -> Tuple[FeatureArray, float]:
        return (self._default_state_definition(_id), 0.0)

    def _extract_sub_components(self) -> Dict[str, SubComponentInfo]:
        '''
        Extract all sub components.

        Notes
        -----
        Each sub component is a method that computes the value of the given
        sub component. The method should have the following properties:
        * Method's name should start with "_sub_comp_".
        * The first argument (except for `self`) should be `_id` which is the
          ID of the object using this sub component.
        * Method should have `**kwargs` argument to avoid raising exceptions if
          unnecessary arguments are passed on to it.
        * Method can have arguments with default values
        * Method should return a dictionary with mandatory keys `name` and
          `value` and optional keys, such as `lower` and `upper`, and
          `categories`.

        Example
        -------
        >>> class Dummy(Stateful):
        ...     def __init__(self) -> None:
        ...         self._some_attribute = 'some value'
        ...         sub_comp_list = self._extract_sub_components()
        ...         self.a_component = Component(tuple(sub_comp_list))
        ...
        ...     def _sub_comp_01(self, _id, **kwargs):
        ...         return {'name': 'sub_comp_01', 'value': 'something'}
        ...
        ...     def _sub_comp_02(self, _id, arg_01, **kwargs):
        ...         return {'name': 'sub_comp_02',
        ...                 'value': self._some_attribute * arg_01}
        >>> d = Dummy()
        >>> d.a_component.add_definition(
        ...     'a_definition',
        ...     (SubComponentInstance('01'),
        ...      SubComponentInstance('02', {'arg_01': 3})))
        >>> print(d.a_component('a_definition', _id=1).value)
        {'sub_comp_01': 'something', 'sub_comp_02':
        'some valuesome valuesome value'}
        >>> d._some_attribute = 'new value'
        >>> print(d.a_component('a_definition', _id=1).value)
        {'sub_comp_01': 'something', 'sub_comp_02':
        'new valuenew valuenew value'}
        '''
        sub_comp_list: Dict[str, SubComponentInfo] = {}
        for func_name, func in ((f, getattr(self, f).__func__)
                                for f in dir(self)
                                if f.startswith('_sub_comp_')):
            if callable(func):
                keywords = list(func.__code__.co_varnames)

                if 'kwargs' in keywords:
                    keywords.remove('kwargs')

                if len(keywords) < 2 or keywords[1] != '_id':
                    raise ValueError(
                        f'Error in {func_name} signature: '
                        'At least two arguments should be accepted. '
                        'The first argument will receive a reference to the '
                        'object (self), and the second argument should be '
                        '"_id", which is the ID of the caller.')

                keywords.remove('_id')

                sub_comp_list[func_name[10:]] = (func, tuple(keywords))

        return sub_comp_list

    def register(self, entity_name: str, _id: Optional[int] = None) -> int:
        '''
        Register an `entity` and return its ID. If the `entity` is new, a new
        ID is generated and the `entity_name` is added to the list of
        registered entities.

        Arguments
        ---------
        entity_name:
            The name of the `entity` to be registered.

        _id:
            The ID of the entity to be used. If not provided, instance will
            assign an ID to the `entity`.

        Returns
        -------
        :
            ID of the registered `entity`.

        Raises
        ------
        ValueError:
            Attempt to register an already registered `entity` with a new ID.

        ValueError:
            Attempt to register an `entity` with an already assigned ID.

        ValueError:
            Reached max capacity.
        '''
        return self._entity_list.append(entity_name=entity_name, _id=_id)

    def deregister(self, entity_id: int) -> None:
        '''
        Deregister an `entity` given its ID.

        Arguments
        ---------
        entity_id:
            The ID of the `entity` to be deregistered.
        '''
        self._entity_list.remove(entity_id)

    def load(self, filename: str,
             path: Optional[Union[str, pathlib.PurePath]]) -> None:

        super().load(filename, path=path)

        self.state.object_ref = self
        self.statistic.set_primary_component(self.state)
        self.state.set_default_definition(
            self._default_state_definition)
        self.statistic.set_default_definition(
            self._default_statistic_definition)

    def save(self,
             filename: Optional[str] = None,
             path: Optional[Union[str, pathlib.PurePath]] = None,
             data_to_save: Optional[Tuple[str, ...]] = None
             ) -> Tuple[pathlib.PurePath, str]:

        object_ref_temp, self.state.object_ref = self.state.object_ref, None
        state_default, self.state._default = self.state._default, None

        prim_comp, self.statistic._primary_component = (  # type: ignore
            self.statistic._primary_component, None)
        statistic_default, self.statistic._default = (
            self.statistic._default, None)

        try:
            f, p = super().save(filename, path=path, data_to_save=data_to_save)
        finally:
            self.state.object_ref = object_ref_temp
            self.state._default = state_default

            self.statistic._primary_component = prim_comp
            self.statistic._default = statistic_default

        return f, p

    def reset(self):
        super().reset()
        self._entity_list.clear()



# -*- coding: utf-8 -*-
'''
ThreePhaseDosingProtocol class
==============================

A dosing protocol class that can contain three dosing protocols for
`initial`, `adjustment` and `maintenance` phases of dosing.
'''

from typing import Any, Dict

from reil.healthcare.dosing_protocols.dosing_protocol import (
    AdditionalInfo, DosingDecision, DosingProtocol)


class ThreePhaseDosingProtocol:
    '''
    A dosing protocol class that can contain three dosing protocols for
    `initial`, `adjustment` and `maintenance` phases of dosing.
    '''

    def __init__(
            self,
            initial_protocol: DosingProtocol,
            adjustment_protocol: DosingProtocol,
            maintenance_protocol: DosingProtocol) -> None:
        '''
        Arguments
        ---------
        initial_protocol:
            A dosing protocol for the initial phase of dosing.

        adjustment_protocol
            A dosing protocol for the adjustment phase of dosing.

        maintenance_protocol
            A dosing protocol for the maintenance phase of dosing.
        '''
        self._initial_protocol = initial_protocol
        self._adjustment_protocol = adjustment_protocol
        self._maintenance_protocol = maintenance_protocol
        self._additional_info: AdditionalInfo = {}

    def prescribe(
            self, patient: Dict[str, Any]) -> DosingDecision:
        '''
        Prescribe a dose and next test (in days) for the given `patient`.

        Arguments
        ---------
        patient:
            A dictionary of patient characteristics necessary to make dosing
            decisions.

        Returns
        -------
        :
            The prescribed dose and the time of the next test (in days).
        '''
        raise NotImplementedError

    def reset(self) -> None:
        '''Reset the dosing protocol.'''
        self._initial_protocol.reset()
        self._adjustment_protocol.reset()
        self._maintenance_protocol.reset()
        self._additional_info = {}



# -*- coding: utf-8 -*-
'''
Buffers module for reinforcement learning
==========================================

This module provides different types of buffer.

Classes
-------
Buffer:
    Base class for all buffers. It provides basic functionality for setup,
    adding to and picking from the buffer.

CircularBuffer:
    A buffer that overwrites the oldest items if more than its size
    is added.

EndlessBuffer:
    A buffer without an end! This buffer keeps appending new items,
    so it should be used with memory and performance constraints in mind.

VanilaExperienceReplay:
    A `CircularBuffer` that only picks randomly and returns
    values only if it is full.


'''

from .buffer import Buffer, PickModes  # noqa: W0611
from .circular_buffer import CircularBuffer  # noqa: W0611
from .endless_buffer import EndlessBuffer  # noqa: W0611
from .sink import Sink  # noqa: W0611
from .vanilla_experience_replay import VanillaExperienceReplay  # noqa: W0611



# -*- coding: utf-8 -*-
'''
WarfarinClusterAgent class
==========================

An agent that produces action based on similarity of the state with clusters.
'''


import os
from dill import HIGHEST_PROTOCOL, dump, load
from random import choice, random
from time import time
from scipy.stats import norm

from collections import deque
import numpy as np
import pandas as pd

from reil.agents import Agent
from reil.datatypes import ReilData


class WarfarinClusterAgent(Agent):
    '''
    An agent that produces action based on similarity of the state with clusters.

    Constructor Arguments
    ---------------------
        default_actions: list of default actions.

    Methods
    -------
        act: return an action based on the given state.
        learn: learn using either history or action, reward, and state.
        reset:
    '''

    def __init__(self, cluster_filename, type='smoothing', smoothing_dose_threshold=0.0, dose_step=0.5, **kwargs):
        '''
        Initialize a WarfarinClusterAgent.
        type:
            smoothed: `smoothing_dose_threshold` and `dose_step` can be set. 
            rule based: argument `rule_base_filename` should specify the file that contains rules.
            two phase: argument `phase_change_day` should specify the change day. During each phase, average dose for the cluster is applied.
        '''
        self._default_actions = {}
        self._cluster_filename=cluster_filename,
        self._type = 'smoothed'
        self._day=0
        self._max_day = 90
        self._previous_dose = 0.0
        self._max_dose = 15.0
        self._smoothing_dose_threshold = 0.0
        self._dose_step = 0.5
        self._rule_base_filename = None
        self._phase_change_day = 0

        self.set_params(**kwargs)
        super().__init__(**kwargs)

        try:
            self._cluster_data = pd.read_csv(self._cluster_filename, sep=',', index_col='Index')
        except FileNotFoundError:
            raise FileNotFoundError(f'{kwargs["cluster_filename"]} not found.')

    def learn(self, **kwargs):
        '''
        This agent does not learn using `learn` method. Instead use ????
        '''
        pass

    def _assign_to_cluster(self, age, CYP2C9, VKORC1):
        prob = {}
        prob['age'] = norm.pdf(age,
                               self._cluster_data.loc['age mean', :],
                               self._cluster_data.loc['age stdev', :])

        prob['CYP2C9'] = self._cluster_data.loc['*1/*1', :].values ** int(CYP2C9 == '*1/*1') \
                       * self._cluster_data.loc['*1/*2', :].values ** int(CYP2C9 == '*1/*2') \
                       * self._cluster_data.loc['*1/*3', :].values ** int(CYP2C9 == '*1/*3') \
                       * self._cluster_data.loc['*2/*2', :].values ** int(CYP2C9 == '*2/*2') \
                       * self._cluster_data.loc['*2/*3', :].values ** int(CYP2C9 == '*2/*3') \
                       / self._cluster_data.loc['CYP2C9 Total', :].values
                       #    * self._cluster_data.loc['*3/*3', :].values ** int(CYP2C9 == '*3/*3') \

        prob['VKORC1'] = self._cluster_data.loc['A/A', :].values ** int(CYP2C9 == 'A/A') \
                       * self._cluster_data.loc['G/A', :].values ** int(CYP2C9 == 'G/A') \
                       * self._cluster_data.loc['G/G', :].values ** int(CYP2C9 == 'G/G') \
                       / self._cluster_data.loc['VKORC1 Total', :].values

        label = np.argmax(prob['age'] * prob['CYP2C9'] * prob['VKORC1']
                        * self._cluster_data.loc['prob', :].values)

        return label

    def act(self, state, **kwargs):
        '''
        return the best action for a given state.

        Arguments
        ---------
            state: the state for which an action is chosen.
        '''
        if self._day >= self._max_day:
            self._day = 0
            self._previous_dose = 0.0

        if self._day == 0:
            self._cluster_label = self._assign_to_cluster(state.normalize().value[0][0], state.value['CYP2C9'], state.value['VKORC1'])

        if self._type == 'smoothed':
            try:
                dose = round(self._cluster_data.at[f'dose {self._day:02} mean', str(self._cluster_label)] / self._dose_step) * self._dose_step
            except ZeroDivisionError:
                dose = self._cluster_data.at[f'dose {self._day:02} mean', str(self._cluster_label)]

            if abs(dose - self._previous_dose) >= self._smoothing_dose_threshold:
                action = ReilData.single_numerical(name='dose', value=dose, lower=0.0, upper=15.0)
                self._previous_dose = dose
            else:
                action = ReilData.single_numerical(
                    name='dose', value=self._previous_dose, lower=0.0, upper=15.0)
        elif self._type == 'two phase':
            if self._day == 0:
                self._dose = np.average(self._cluster_data.loc[[f'dose {i:02} mean' for i in range(self._phase_change_day)],
                                                               str(self._cluster_label)])
            elif self._day == self._phase_change_day:
                self._dose = np.average(self._cluster_data.loc[[f'dose {i:02} mean' for i in range(self._phase_change_day, self._max_day)],
                                                               str(self._cluster_label)])

            action = ReilData.single_numerical(
                name='dose', value=self._dose, lower=0.0, upper=15.0)

        elif self._type == 'rule based':
            if self._day == 0:
                self._rule_base_filename = kwargs.get('rule_base_filename', self._rule_base_filename)
                self._rule_base_data = pd.read_csv(self._rule_base_filename, sep=',', index_col='Index')

        self._day += 1

        return action

    def __repr__(self):
        try:
            return f'WarfarinClusterAgent({self._cluster_filename})'
        except AttributeError:
            return 'WarfarinClusterAgent'

# c = WarfarinClusterAgent(cluster_filename='Weka output (4 clusters).csv')
# [[(age/15, CYP2C9, VKORC1, c._assign_to_cluster(age/15, CYP2C9, VKORC1))] for age in range(15) for CYP2C9 in ('*1/*1', '*1/*2', '*1/*3', '*2/*2', '*2/*3') for VKORC1 in ('G/G', 'G/A', 'A/A')]



# -*- coding: utf-8 -*-
'''
ReilData class
==============

The main datatype used to communicate `state`s, `action`s, and `reward`s,
between objects in `reil`. `ReilData` is basically a tuple that contains
instances of `BaseData`, `CategoricalData`, and `NumericalData`.
'''
from __future__ import annotations

import dataclasses
import itertools
from dataclasses import field
from typing import (Any, Callable, Dict, Generic, Iterable, Iterator, List,
                    Literal, Mapping, Optional, Sequence, Tuple, TypeVar,
                    Union, cast, overload)

T = TypeVar('T')

Categorical = TypeVar('Categorical')
Numerical = TypeVar('Numerical', int, float)

Normal = Union[Literal[0], Literal[1], float]
Normalized = Union[Normal, Sequence[Normal], None]
Normalizer = Callable[..., Normalized]


@dataclasses.dataclass(frozen=True)
class BaseData(Generic[T]):
    '''
    The base class for `ReilData` elements.

    Attributes
    ----------
    name:
        Name of the data.

    value:
        Value of the data. Can be one item or a sequence.

    normalizer:
        A function that gets a `BaseData` object and returns its normalized
        form.

    lazy_evaluation:
        Whether normalization should be done at the point of instantiation
        (False) or at the access time (True).


    :meta private:
    '''
    name: str
    value: Optional[Union[T, Sequence[T]]] = None
    normalizer: Optional[Normalizer] = lambda _: None
    lazy_evaluation: Optional[bool] = field(default=None, compare=False)
    is_numerical: Optional[bool] = field(
        default=None, init=False, repr=False, compare=False)
    _normalized: Normalized = field(
        default=None, init=False, repr=False, compare=False)

    def __post_init__(self):
        # if not isinstance(self.value, Hashable):
        #     warnings.warn('Non-hashable object!')
        self._validate(self)
        if self.normalizer is None:
            object.__setattr__(self, 'normalizer',
                               self._default_normalizer)
        if not self.lazy_evaluation:
            object.__setattr__(self, '_normalized',
                               self.normalizer(self))

    @staticmethod
    def _validate(data: BaseData[T]) -> None:
        '''
        Validate the input data.

        Arguments
        ---------
        data:
            An instance to be validated.

        Raises
        ------
        ValueError:
            if `data.value` does not match the validation criteria.


        :meta public:
        '''
        pass

    @staticmethod
    def _default_normalizer(data: BaseData[T]) -> Normalized:
        '''
        Normalize the data.

        Arguments
        ---------
        data:
            An instance to be normalized.

        Returns
        -------
        :
            The normalized form of `data.value`.
        '''
        return None

    @property
    def normalized(self) -> Normalized:
        if self._normalized is None:
            object.__setattr__(self, '_normalized',
                               self.normalizer(self))

        return self._normalized

    def as_dict(self) -> Dict[str, Any]:
        '''
        Return the data as a dictionary.

        Returns
        -------
        :
            The data as a dictionary.
        '''
        return {'name': self.name, 'value': self.value}


@dataclasses.dataclass(frozen=True)
class CategoricalData(BaseData[Categorical]):
    '''
    A datatype for categorical data.

    Attributes
    ----------
    name:
        Name of the data.

    value:
        Value of the data. Can be one item or a sequence.

    normalizer:
        A function that gets a `BaseData` object and returns its normalized
        form.

    lazy_evaluation:
        Whether normalization should be done at the point of instantiation
        (False) or at the access time (True).

    categories:
        A list of all categories.
    '''
    categories: Optional[Tuple[Categorical, ...]] = None

    def __post_init__(self):
        super().__post_init__()
        object.__setattr__(self, 'is_numerical', False)

    @staticmethod
    def _default_normalizer(data: CategoricalData) -> Normalized:
        value = data.value
        categories = data.categories
        if categories is None or value is None:
            return None

        if isinstance(value, type(categories[0])):
            return tuple(int(x_i == value)
                         for x_i in categories)

        return tuple(int(x_i == v)
                     for v in value
                     for x_i in categories)

    @staticmethod
    def _validate(data: CategoricalData) -> None:
        value = data.value
        if value is not None:
            is_sequence = isinstance(value, (list, tuple))

            categories = data.categories
            if categories is not None:
                if ((not is_sequence and value not in categories) or
                        (is_sequence and any(v not in categories
                                             for v in value))):
                    raise ValueError(
                        f'value={value} is '
                        f'in the categories={categories}.')

    def as_dict(self) -> Dict[str, Any]:
        return {'name': self.name, 'value': self.value,
                'categories': self.categories,
                'is_numerical': self.is_numerical}


@dataclasses.dataclass(frozen=True)
class NumericalData(BaseData[Numerical]):
    '''
    A datatype for numerical data.

    Attributes
    ----------
    name:
        Name of the data.

    value:
        Value of the data. Can be one item or a sequence.

    normalizer:
        A function that gets a `BaseData` object and returns its normalized
        form.

    lazy_evaluation:
        Whether normalization should be done at the point of instantiation
        (False) or at the access time (True).

    lower:
        The lower bound of value.

    upper:
        The upper bound of value.
    '''
    lower: Optional[Numerical] = None
    upper: Optional[Numerical] = None

    def __post_init__(self):
        super().__post_init__()
        object.__setattr__(self, 'is_numerical', True)

    @staticmethod
    def _default_normalizer(data: NumericalData) -> Normalized:
        value = data.value
        lower = data.lower
        upper = data.upper
        if value is None or lower is None or upper is None:
            return None

        denominator = upper - lower

        sequence_check = isinstance(value, (list, tuple))
        try:
            if sequence_check:
                return tuple((v - lower) / denominator
                             for v in value)
            else:
                return (value - lower) / denominator
        except ZeroDivisionError:
            return [1] * len(value) if sequence_check else 1

    @staticmethod
    def _validate(data: NumericalData) -> None:
        value = data.value
        lower = data.lower
        upper = data.upper
        if value is not None:
            is_sequence = isinstance(value, (list, tuple))

            if lower is not None:
                if ((not is_sequence and value < lower) or
                        (is_sequence and any(v < lower
                                             for v in value))):
                    raise ValueError(
                        f'value={value} is less than lower={lower}.')

            if upper is not None:
                if ((not is_sequence and value > upper) or
                        (is_sequence and any(v > upper
                                             for v in value))):
                    raise ValueError(
                        f'value={value} is '
                        f'greater than upper={upper}.')

    def as_dict(self) -> Dict[str, Any]:
        return {'name': self.name, 'value': self.value,
                'lower': self.lower, 'upper': self.upper,
                'is_numerical': self.is_numerical}


ReilDataInput = Union[Mapping[str, Any],
                      BaseData[Any],
                      CategoricalData[Categorical],
                      NumericalData[Numerical]]


class ReilData(Sequence[ReilDataInput], Generic[Categorical, Numerical]):
    '''
    The main datatype used to communicate `state`s, `action`s, and `reward`s,
    between objects in `reil`.
    '''

    def __init__(self,
                 data: Union[ReilDataInput,
                             Sequence[ReilDataInput],
                             Iterator[ReilDataInput]],
                 lazy_evaluation: Optional[bool] = None):
        '''
        Arguments
        ---------
        data:
            One or a sequence of `BaseData`, `CategoricalData`,
            `NumericalData`, or dicts that include 'name'. Other attributes are
            optional. If none of `categories`, `lower` and `upper` are
            provided, the object is assumed `BaseData`.

        lazy_evaluation:
            Whether to compute normalized value at the time of instantiation or
            at the time of the first access. If not provided, `ReilData` looks
            for `lazy_evaluation` in each item. If failed, True is assumed.
        '''
        temp: List[Union[BaseData[Any],
                         CategoricalData[Categorical],
                         NumericalData[Numerical]]] = []
        _data = data if isinstance(data, (Sequence, Iterator)) else [data]
        for d in _data:
            if isinstance(d, BaseData):
                temp.append(d)
            elif isinstance(d, dict):
                name = d['name']
                value = d.get('value')
                normalizer = d.get('normalizer')
                l_e = lazy_evaluation if lazy_evaluation is not None \
                    else d.get('lazy_evaluation', True)

                if 'categories' in d:
                    temp.append(CategoricalData(
                        name=name,
                        value=value,
                        categories=d['categories'],
                        normalizer=normalizer,
                        lazy_evaluation=l_e))
                elif 'lower' in d or 'upper' in d:
                    temp.append(NumericalData(
                        name=name,
                        value=value,
                        lower=d.get('lower'),
                        upper=d.get('upper'),
                        normalizer=normalizer,
                        lazy_evaluation=l_e))
                else:
                    temp.append(BaseData(
                        name=name,
                        value=value,
                        normalizer=normalizer,
                        lazy_evaluation=l_e))
            else:
                raise TypeError(f'Unknow input type {type(d)} for item: {d}')

        self._data = tuple(temp)
        self._clear_temps()

    @classmethod
    def single_base(
            cls,
            name: str,
            value: Optional[Union[T, Sequence[T]]] = None,
            normalizer: Optional[Normalizer] = None,
            lazy_evaluation: Optional[bool] = None) -> ReilData:
        '''
        Create a `ReilData` instance.

        Arguments
        ---------
        name:
            Name of the instance.

        value:
            The value to store.

        normalizer:
            A function that accepts a `BaseData` object and returns the
            normalized value.

        lazy_evaluation:
            Whether normalization should be done at the point of instantiation
            (False) or at the access time (True).
        '''
        instance = cls([])
        instance._data = (
            BaseData(name=name,
                     value=value,
                     normalizer=normalizer,
                     lazy_evaluation=lazy_evaluation),)

        return instance

    @classmethod
    def single_categorical(
            cls,
            name: str,
            value: Optional[Union[Categorical, Sequence[Categorical]]] = None,
            normalizer: Optional[Normalizer] = None,
            lazy_evaluation: Optional[bool] = None,
            categories: Optional[Tuple[Categorical, ...]] = None) -> ReilData:
        '''
        Create a `ReilData` instance.

        Arguments
        ---------
        name:
            Name of the instance.

        value:
            The value to store.

        categories:
            A list of categories if the object is categorical

        normalizer:
            A function that accepts a `BaseData` object and returns the
            normalized value.

        lazy_evaluation:
            Whether normalization should be done at the point of instantiation
            (False) or at the access time (True).
        '''
        instance = cls([])
        instance._data = (
            CategoricalData(
                name=name,
                value=value,
                categories=categories,
                normalizer=normalizer,
                lazy_evaluation=lazy_evaluation),)

        return instance

    @classmethod
    def single_numerical(
            cls,
            name: str,
            value: Optional[Union[Numerical, Sequence[Numerical]]] = None,
            normalizer: Optional[Normalizer] = None,
            lazy_evaluation: Optional[bool] = None,
            lower: Optional[Numerical] = None,
            upper: Optional[Numerical] = None) -> ReilData:
        '''
        Create a `ReilData` instance.

        Arguments
        ---------
        name:
            Name of the instance.

        value:
            The value to store.

        lower:
            The lower bound of value if data is numerical

        upper:
            The upper bound of the value is numerical

        normalizer:
            A function that accepts a `BaseData` object and returns the
            normalized value.

        lazy_evaluation:
            Whether normalization should be done at the point of instantiation
            (False) or at the access time (True).
        '''
        instance = cls([])
        instance._data = (
            NumericalData(
                name=name,
                value=value,
                lower=lower,
                upper=upper,
                normalizer=normalizer,
                lazy_evaluation=lazy_evaluation),)

        return instance

    @classmethod
    def single_item(
            cls,
            name: str,
            value: Optional[Union[
                Numerical, Sequence[Numerical],
                Categorical, Sequence[Categorical]]] = None,
            normalizer: Optional[Normalizer] = None,
            lazy_evaluation: Optional[bool] = None,
            categories: Optional[Tuple[Categorical, ...]] = None,
            lower: Optional[Numerical] = None,
            upper: Optional[Numerical] = None) -> ReilData:
        '''
        Create a `ReilData` instance.

        Arguments
        ---------
        name:
            Name of the instance.

        value:
            The value to store.

        categories:
            A list of categories if the object is categorical

        lower:
            The lower bound of value if data is numerical

        upper:
            The upper bound of the value is numerical

        normalizer:
            A function that accepts a `BaseData` object and returns the
            normalized value.

        lazy_evaluation:
            Whether normalization should be done at the point of instantiation
            (False) or at the access time (True).
        '''
        instance = cls([])
        if lower is not None or upper is not None:
            instance._data = (
                NumericalData(
                    name=name,
                    value=value,  # type: ignore
                    lower=lower,
                    upper=upper,
                    normalizer=normalizer,
                    lazy_evaluation=lazy_evaluation),)
        elif categories is not None:
            instance._data = (
                CategoricalData(
                    name=name,
                    value=value,
                    categories=categories,
                    normalizer=normalizer,
                    lazy_evaluation=lazy_evaluation),)
        else:
            instance._data = (
                CategoricalData(
                    name=name,
                    value=value,
                    categories=categories,
                    normalizer=normalizer,
                    lazy_evaluation=lazy_evaluation),)

        return instance

    def _clear_temps(self):
        self._value: Optional[Dict[str, Any]] = None
        self._lower: Optional[Dict[str, Numerical]] = None
        self._upper: Optional[Dict[str, Numerical]] = None
        self._categories: Optional[Dict[str, Tuple[Categorical, ...]]] = None
        self._is_numerical: Optional[Dict[str, bool]] = None

    def index(self, value: Any,
              start: int = 0, stop: Optional[int] = None) -> int:
        _stop = stop if stop is not None else len(self._data)
        if isinstance(value, BaseData):
            for i in range(start, _stop):
                if self._data[i] == value:
                    return i

            raise ValueError(f'{value} is not on the list.')

        elif isinstance(value, type(self._data[0].name)):
            for i in range(start, _stop):
                if self._data[i].name == value:
                    return i

            raise ValueError(f'{value} is not on the list.')

        else:
            raise ValueError(f'{value} is not on the list.')

    @property
    def value(self) -> Dict[str, Any]:
        '''
        Return a dictionary with elements' names as keys and
        their respective values as values.

        Returns
        -------
        :
            Names of the elements and their values.
        '''
        if self._value is None:
            self._value = {v.name: v.value
                           for v in self._data}

        return self._value

    @property
    def lower(self) -> Dict[str, Numerical]:
        '''
        Return all `lower` attributes.

        Returns
        -------
        :
            `lower` attribute of all `NumericalData` variables with their names
            as keys.
        '''
        if self._lower is None:
            self._lower = {v.name: cast(Numerical, v.lower)  # type: ignore
                           for v in self._data
                           if hasattr(v, 'lower')}

        return self._lower

    @property
    def upper(self) -> Dict[str, Numerical]:
        '''
        Return all `upper` attributes.

        Returns
        -------
        :
            `upper` attribute of all `NumericalData` variables with their names
            as keys.
        '''
        if self._upper is None:
            self._upper = {v.name: cast(Numerical, v.upper)  # type: ignore
                           for v in self._data
                           if hasattr(v, 'upper')}

        return self._upper

    @property
    def categories(self) -> Dict[str, Tuple[Categorical, ...]]:
        '''
        Return all `categories` attributes.

        Returns
        -------
        :
            `categories` attribute of all `CategoricalData` variables with
            their names as keys.
        '''
        if self._categories is None:
            self._categories = {v.name:
                                cast(Tuple[Categorical, ...],
                                     v.categories)  # type: ignore
                                for v in self._data
                                if hasattr(v, 'categories')}

        return self._categories

    @property
    def normalized(self) -> ReilData:
        '''
        Normalize all items in the instance.

        Returns
        -------
        :
            A `ReilData` consist of normalized values of all the items in the
            instance, in the form of `NumericalData` objects.
        '''
        return ReilData(
            NumericalData(name=v.name, value=v.normalized,
                          lower=0, upper=1, lazy_evaluation=True)
            for v in self._data)

    def flatten(self) -> List[BaseData]:
        """Combine values of all items in the instance.

        Returns
        -------
        :
            A list that contains all the values of all the items.
        """
        def make_iterable(x: T) -> Iterable[T]:
            return x if isinstance(x, Iterable) else [x]

        return list(itertools.chain(*[make_iterable(sublist)
                                      for sublist in self.value.values()]))

    def split(self) -> Union[ReilData, List[ReilData]]:
        """Split the `ReilData` into a list of `ReilData`.

        Returns
        -------
        :
            All items in the instance as separate `ReilData` instances.
        """
        if len(self) == 1:
            d = self._data[0]
            if not isinstance(d.value, (list, tuple)):
                splitted_list = ReilData(self._data)
            else:
                temp = d.as_dict()
                cls = type(d)
                value = temp['value']
                del temp['value']
                if 'is_numerical' in temp:
                    del temp['is_numerical']

                splitted_list = [
                    ReilData(cls(
                        value=v,
                        **temp))
                    for v in value]

        else:
            splitted_list = cast(List[ReilData], list(self._data))

        return splitted_list

    @overload
    def __getitem__(self, i: int) -> BaseData:
        ...

    @overload
    def __getitem__(self, s: slice) -> Tuple[BaseData, ...]:
        ...

    def __getitem__(self, x):
        return self._data.__getitem__(x)

    def __len__(self) -> int:
        return self._data.__len__()

    def __hash__(self) -> int:
        return hash(tuple(self._data))

    def __eq__(self, other) -> bool:
        return isinstance(other, type(self)) and (self._data == other._data)

    def __add__(self, other: Union[BaseData[T], ReilData]) -> ReilData:
        if isinstance(other, BaseData):
            if other.name in self._data:
                raise ValueError(
                    'Cannot have items with same names.'
                    ' Use update() if you need to update an item.')
            new_data = [other]
        elif isinstance(other, ReilData):
            for k in other:
                if k in self._data:
                    raise ValueError(
                        'Cannot have items with same names.'
                        ' Use update() if you need to update an item.')
            new_data = other._data
        else:
            raise TypeError(
                'Concatenation of type ReilData'
                f' and {type(other)} not implemented!')

        return ReilData(itertools.chain(self._data, new_data))

    def __neg__(self) -> ReilData:
        temp = [v.as_dict()
                for v in self._data]
        for item in temp:
            item['value'] = -item['value']

        return ReilData(temp)

    def __repr__(self) -> str:
        return f'[{super().__repr__()} -> {self._data}]'

    def __str__(self) -> str:
        return f"[{', '.join((d.__str__() for d in self._data))}]"


if __name__ == "__main__":
    # from timeit import timeit

    def f1():
        t1 = ReilData.single_categorical(name='test A', value=['a', 'b'],
                                         categories=('a', 'b'))
        t1 = ReilData.single_numerical(name='test A', value=[1, 2])
        t1 = ReilData.single_base(name='test A', value=['a', 'b'])
        return t1

    def f2():
        t2 = ReilData(({'name': x, 'value': x,
                        'categories': list('abcdefghijklmnopqrstuvwxyz')}
                       for x in 'abcdefghijklmnopqrstuvwxyz'),
                      lazy_evaluation=True)
        return t2.normalized.flatten()

    def f3():
        t2 = ReilData([{'name': 'A', 'value': 'a', 'categories': ['a', 'b']},
                       {'name': 'B', 'value': [10, 20],
                        'lower': 0, 'upper': 100}], lazy_evaluation=True)
        return t2

    # t1 = f1()
    # t2 = ReilData(t1)
    print(f1() + f3())
    # test = f1().split()
    # print(test)
    # print(timeit(f1, number=1000))
    # print(timeit(f2, number=100))
    # print(timeit(f3, number=1000))
    # print(f2().normalized.as_list())
    # print(a.values)
    # print(a.lower)
    # print(a.categories)
    # a.values = {'test A':'b'}



# -*- coding: utf-8 -*-
'''
functions module
================

Contains some useful functions.
'''

import math
import random
from typing import Any, Callable, Iterable, List, Optional, Tuple, TypeVar

import numpy as np
from reil.datatypes.feature import FeatureArray, FeatureGenerator
from scipy.stats import lognorm


def random_choice(f: Any):
    '''
    This function allows `yaml` config files to use `random.choice`
    as part of `reil` module.
    '''
    return random.choice(f)


def random_uniform(f: FeatureGenerator[float]) -> float:
    if f.randomized:
        return np.random.uniform(f.lower, f.upper)

    if f.mean is not None:
        return f.mean

    if f.upper is None or f.lower is None:
        raise ValueError('mean, or upper and lower should be numbers.')

    return (f.upper - f.lower) / 2.0


def random_normal(f: FeatureGenerator[float]) -> float:
    if f.randomized:
        return np.random.normal(f.mean, f.stdev)

    if f.mean is None:
        raise ValueError('mean should be a number.')

    return f.mean


def random_normal_truncated(f: FeatureGenerator[float]) -> float:
    if f.randomized:
        return min(max(
            np.random.normal(f.mean, f.stdev), f.lower),
            f.upper)

    if f.mean is None:
        raise ValueError('mean should be a number.')

    return f.mean


def random_lognormal(f: FeatureGenerator[float]) -> float:
    try:
        exp_mu = math.exp(f.mean)  # type: ignore
    except TypeError:
        raise ValueError('mean should be a number.')

    if f.randomized:
        return lognorm.rvs(s=f.stdev, scale=exp_mu)

    return exp_mu


def random_lognormal_truncated(f: FeatureGenerator[float]) -> float:
    # capture 50% of the data.
    # This restricts the log values to a "reasonable" range
    try:
        exp_mu = math.exp(f.mean)  # type: ignore
    except TypeError:
        raise ValueError('mean should be a number.')

    if f.randomized:
        quartileRange = (0.25, 0.75)
        lnorm = lognorm(f.stdev, scale=exp_mu)
        qValues: Tuple[float, float] = lnorm.ppf(quartileRange)
        values: List[float] = list(
            v for v in lnorm.rvs(size=1000)
            if (v > qValues[0]) & (v < qValues[1]))

        return random.sample(values, 1)[0]

    return exp_mu


Categorical = TypeVar('Categorical')


def random_categorical(f: FeatureGenerator[Categorical]) -> Categorical:
    if (categories := f.categories) is None:
        raise TypeError('No categories found!')

    if f.randomized:
        if (probs := f.probabilities) is None:
            return random.choice(categories)
        else:
            return np.random.choice(categories, 1, p=probs)[0]

    return categories[0]


def square_dist(x: float, y: Iterable[float]) -> float:
    return sum((x - yi) ** 2
               for yi in y)


def in_range(r: Tuple[float, float], x: Iterable[float]) -> int:
    return sum(r[0] <= xi <= r[1]
               for xi in x)


def interpolate(start: float, end: float, steps: int) -> Iterable[float]:
    return (start + (end - start) / steps * j
            for j in range(1, steps + 1))


T = TypeVar('T')


def generate_modifier(
        operation: Callable[[T], T],
        condition: Optional[Callable[[FeatureArray], bool]] = None
) -> Callable[[FeatureArray, T], T]:
    '''Generate a modifier function for states or actions

    Parameters
    ----------
    operation:
        What should happen to the input.

    condition:
        A function that accepts a state `FeatureArray`, and based on that
        determines if the `operation` should be applied to the input.

    Returns
    -------
    :
        A function that accepts `condition_state` and `input` and returns the
        modified `input`.
    '''
    if condition is None:
        def no_condition_modifier(
                condition_state: FeatureArray, input: T) -> T:
            return operation(input)

        return no_condition_modifier
    else:
        def modifier(
                condition_state: FeatureArray, input: T) -> T:
            if condition(condition_state):  # type: ignore
                return operation(input)

            return input

    return modifier



import pathlib
from typing import Dict, List, Optional, Union

from reil.environments.session import Session
from reil.environments.task import Task
from reil.utils import ConfigParser, OutputWriter

# TODO: documentation


class SessionBuilder:
    def __init__(
            self, config_filenames: Dict[str, str],
            config_path: Optional[Union[pathlib.Path, str]],
            vars_dict: Optional[Dict[str, str]]) -> None:
        self._parser = ConfigParser(
            config_filenames=config_filenames, config_path=config_path,
            vars_dict=vars_dict)

    def create_task(
            self,
            task_name: str,
            parent_session_path: pathlib.PurePath = pathlib.PurePath('.')
    ) -> Task:
        components = self._parser.extract('tasks', task_name)

        name = components['name']
        path = components['path']
        agent_training_triggers = components['agent_training_triggers']

        interaction_sequence = tuple(
            self._parser.extract('interaction_protocols', name, as_object=True)
            for name in components['interaction_sequence']
        )

        start_iteration = components.get('start_iteration', 0)
        max_iterations = components.get('max_iterations', 1)
        save_iterations = components.get('save_iterations', True)

        temp = components.get(
            'writer',
            {'name': None, 'path': None, 'columns': None})
        writer = OutputWriter(
            filename=temp.get('name') or name,
            path=pathlib.PurePath(temp.get('path') or
                                  (parent_session_path / 'reports')),
            columns=temp.get('columns'))

        t = Task(
            name=name,
            path=pathlib.PurePath(parent_session_path, path),
            agent_training_triggers=agent_training_triggers,
            interaction_sequence=interaction_sequence,
            start_iteration=start_iteration,
            max_iterations=max_iterations,
            writer=writer,
            save_iterations=save_iterations
        )

        return t

    def create_session(
            self,
            session_name: str,
            parent_session_path: Union[pathlib.PurePath, str] = '.'
    ) -> Session:

        try:
            session_info = self._parser.extract('sessions', session_name)
        except KeyError:
            raise ValueError(f'Session {session_name} not found.')

        components: Dict[str, List[Task]] = {
            'main_task': [],
            'tasks_before': [],
            'tasks_after': [],
            'tasks_before_iteration': [],
            'tasks_after_iteration': []
        }

        for component in components:
            if task_list := session_info.get(component):
                if component == 'main_task':
                    task_list = [task_list]
                for t in task_list:
                    if self._parser.contains('tasks', t):
                        temp = self.create_task(t, pathlib.PurePath(
                            session_info['path'], session_name))
                    # elif self._parser.contains('sessions', t):
                    #     temp = self.create_session(t, pathlib.PurePath(
                    #         session_info['path'], session_name))
                    else:
                        raise ValueError(f'Unknown name "{t}" in {component}.')

                    components[component].append(temp)

        if components['main_task']:
            components['main_task'] = components['main_task'][0]
        else:
            raise ValueError(
                f'main_task cannot be None for session {session_info}.')

        agents = {
            name: self._parser.extract('agents', typ, as_object=True)
            for name, typ in session_info['agents'].items()
        }

        subjects = {
            name: self._parser.extract('subjects', typ, as_object=True)
            for name, typ in session_info['subjects'].items()
        }

        demons = {
            name: self._parser.extract('demons', typ, as_object=True)
            for name, typ in (session_info.get('demons') or {}).items()
        }

        return Session(
            name=session_info['name'],
            path=pathlib.PurePath(
                parent_session_path or '.', session_info['path']),
            agents=agents, subjects=subjects, demons=demons,
            separate_process=session_info.get('separate_process'),
            process_type=session_info.get('process_type'),
            **components)  # type: ignore



# -*- coding: utf-8 -*-
'''
Patient class
=============

This class is the base class to model patients with different characteristics.
'''

from typing import Any, Dict

from reil.datatypes.feature import Feature, FeatureGenerator
from reil.healthcare.mathematical_models import HealthMathModel


class Patient:
    '''
    Base class for patients in healthcare.
    '''
    def __init__(self, model: HealthMathModel, **feature_values: Any) -> None:
        '''
        Parameters
        ----------
        model:
            A `HealthMathModel` to be used to model patient's behavior.

        feature_values:
            Keyword arguments by which some of the `features` of the patient
            can be determined. For example, if "age" is one of the features,
            age=40.0 will set the initial age to 40.0.
        '''
        if not hasattr(self, 'feature_gen_set'):
            self.feature_gen_set: Dict[str, FeatureGenerator[Any]] = {}
        if not hasattr(self, 'feature_set'):
            self.feature_set: Dict[str, Feature[Any]] = {}

        for k in self.feature_gen_set:
            self.feature_set[k] = self.feature_gen_set[k](
                feature_values.get(k))

        self._model = model
        self._model.setup(**self.feature_set)

    def generate(self) -> None:
        '''
        Generate a new patient.

        This method calls every `feature`, and then sets
        up to `model` using the new values.
        '''
        for k in self.feature_gen_set:
            self.feature_set[k] = self.feature_gen_set[k]()

        self._model.setup(**self.feature_set)

    def model(self, **inputs: Any) -> Dict[str, Any]:
        '''Model patient's behavior.

        Arguments
        ---------
        inputs:
            Keyword arguments that specify inputs to the model. For example, if
            `dose` is a necessary input, `model(dose=10.0)` will provide the
            model with dose of 10.0.

        Returns
        -------
        :
            All the outputs of running the mathematical model, given the input.
        '''
        return self._model.run(**inputs)
